/* cplxfy.h                                                     12/28/00

     Copyright (c) John R. Berryhill 1999 - 2000.  All Rights Reserved.

*/
#ifndef CPLXFY_H_INCLUDED
#define CPLXFY_H_INCLUDED

class Csum;
class Cdif;

template<class T> struct cplxfy
{
   T *re, *im;
   cplxfy (T& a, T& b){ re = &a; im = &b; }
   cplxfy (unsigned);
   cplxfy ();

   cplxfy<T>& operator=  (const cplxfy<T>& );
   cplxfy<T>& operator+= (const cplxfy<T>& );
   cplxfy<T>& operator-= (const cplxfy<T>& );
   cplxfy<T>& operator*= (const cplxfy<double>& );
   cplxfy<T>& operator*= (const double& );
   cplxfy<real_vector>& fft(const real_vector& );
   cplxfy<real_vector>& phase_shift(int, unsigned);
   cplxfy<real_vector>& bandpass(double, double, unsigned);
   cplxfy<real_vector>& operator= (const Csum& );
   cplxfy<real_vector>& operator= (const Cdif& );
   cplxfy<real_vector>& operator*= (const cplxfy<matrix>& );
   cplxfy<real_vector>& operator*= (const real_vector& );
   cplxfy<real_vector>& operator*= (const cplxfy<real_vector>& );
};
void cfft( cplxfy<real_vector>&, unsigned, int);
void cfft( cplxfy<matrix>&, unsigned, int);
void cfft( cplxfy<real_vector>*, unsigned, int);
void twofer( cplxfy<real_vector>&, unsigned, int);
void phase_shift_2D( cplxfy<matrix>&, int);
unsigned bitrev( const unsigned, const unsigned);
unsigned next_pow_2(const unsigned);
unsigned sizeof_fft(const real_vector&);
unsigned sizeof_fft(const cplxfy<real_vector>&);

template<class T> inline T& real( const cplxfy<T>& x){ return *(x.re); }

template<class T> inline T& imag( const cplxfy<T>& x){ return *(x.im); }

template<class T> inline void swap(cplxfy<T>& x1, cplxfy<T>& x2)
{
   T *t;  t = x1.re;  x1.re = x2.re;  x2.re = t; //note T is pointer
          t = x1.im;  x1.im = x2.im;  x2.im = t;  return;
}

template<class T> double abs( const cplxfy<T>&);

inline cplxfy<real_vector>::cplxfy (unsigned n)
{
   re = new real_vector(n);  im = new real_vector(n);
}

inline cplxfy<real_vector>::cplxfy ()
{
   re = new real_vector();  im = new real_vector();
}

inline cplxfy<double>&
cplxfy<double>::operator= (const cplxfy<double>& rhs)
{
   *re = *rhs.re; *im = *rhs.im;  return *this;
}

inline cplxfy<real_vector>&
cplxfy<real_vector>::operator= (const cplxfy<real_vector>& rhs)
{
   *re = *rhs.re; *im = *rhs.im;  return *this;
}

inline cplxfy<double>&
cplxfy<double>::operator+= (const cplxfy<double>& rhs)
{
   *re += *rhs.re; *im += *rhs.im; return *this;
}

inline cplxfy<real_vector>&
cplxfy<real_vector>::operator+= (const cplxfy<real_vector>& rhs)
{
   *re += *rhs.re; *im += *rhs.im; return *this;
}

inline cplxfy<double>&
cplxfy<double>::operator-= (const cplxfy<double>& rhs)
{
   *re -= *rhs.re; *im -= *rhs.im; return *this;
}

inline cplxfy<real_vector>&
cplxfy<real_vector>::operator-= (const cplxfy<real_vector>& rhs)
{
   *re -= *rhs.re; *im -= *rhs.im; return *this;
}

inline cplxfy<double>&
cplxfy<double>::operator*= (const double& c)
{
   *re *= c; *im *= c; return *this;
}

inline cplxfy<real_vector>&
cplxfy<real_vector>::operator*= (const double& c)
{//(z = z*a)
   *re *= c; *im *= c; return *this;
}

inline cplxfy<double>&
cplxfy<double>::operator*= (const cplxfy<double>& c)
{
    double tmp;  tmp = (*re)*(*c.re) - (*im)*(*c.im);
    *im = (*re)*(*c.im) + (*im)*(*c.re);  *re = tmp;
    return *this;
}

inline int log2_n( const unsigned length)
{
   return log( (double)length-0.5 ) / log(2.0) + 1.0;
}

struct Csum
{
   cplxfy<real_vector> *v1, *v2;  public:

   Csum( cplxfy<real_vector>& a,
         cplxfy<real_vector>& b){ v1 = &a; v2 = &b; }
};

temp_real_vector real(Csum&);

temp_real_vector imag(Csum&);

inline Csum operator+( cplxfy<real_vector>& a,
                       cplxfy<real_vector>& b){ return Csum(a,b); }

struct Cdif
{
   cplxfy<real_vector> *v1, *v2;

   Cdif( cplxfy<real_vector>& a,
         cplxfy<real_vector>& b){ v1 = &a; v2 = &b; }
};

temp_real_vector real(Cdif&);

temp_real_vector imag(Cdif&);

inline Cdif operator-( cplxfy<real_vector>& a,
                       cplxfy<real_vector>& b){ return Cdif(a,b); }
#endif CPLXFY_H

