/* matrix.h                                                     12/28/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/
#ifndef MATRIX_H_INCLUDED
#define MATRIX_H_INCLUDED
#include "..\book\source\vector.h"
struct pivot;

class matrix
{
   friend class submatrix;

   protected:

   real_vector *v; unsigned mrows, ncols, size;

   public:

   matrix();
   matrix(unsigned, unsigned);
   matrix(real_vector*, unsigned, unsigned);
   matrix(double*, unsigned, unsigned);

   real_vector& operator [] (unsigned) const;
   double& operator () (unsigned, unsigned) const;   // Fortran style
   unsigned M() const { return mrows; }
   unsigned N() const { return ncols; }
   matrix& plus_outer (const real_vector&, const real_vector&);
   void swaprows (const unsigned, const unsigned);
   matrix& row_house(const real_vector&);
   matrix& col_house(const real_vector&);
   matrix& HouseholderQR();
   matrix& HouseBidiag();
   matrix& HouseTridiag();
   unsigned col_rank(double);
   void dumpall () const;
};

unsigned Jacobi(real_vector&, real_vector&, real_vector&, real_vector&);
void SVD(matrix&, matrix&, real_vector&);
void givens_2x2_prepost  (matrix&, unsigned, double, double);
void givens_2x2_symmetric(matrix&, unsigned, double, double);
void alg_822 (matrix&);
void alg_823 (matrix&);
unsigned Crout_LU (matrix&, pivot&);

class submatrix : public matrix
{

public:

   submatrix(const matrix&, unsigned, unsigned);
   submatrix(const real_vector&, unsigned);
   ~submatrix();
   submatrix& operator--();
};

inline matrix::matrix()
{
   v = NULL;  mrows = 0;  ncols = 0;  size = 0;
}

inline matrix::matrix(real_vector* data, unsigned i, unsigned j)
{
   v = data;  mrows = i;  ncols = j;  size = 0;
}

inline real_vector& matrix::operator [] (unsigned i) const
{
   return v[i];
}

inline double& matrix::operator () (unsigned i, unsigned j) const
{
   return v[i-1](j);
}

#endif MATRIX_H
