/* method.cpp                                                   12/28/00

     Copyright (c) John R. Berryhill 2000.  All Rights Reserved.

*/
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include "..\book\source\method.h"
#include "..\book\source\vector.h"
#include "..\book\source\matrix.h"
#include "..\book\source\sparse.h"

#include <string.h>

#ifndef MINMAX_DEFINED
#  define max(a,b)    (((a) > (b)) ? (a) : (b))
#  define min(a,b)    (((a) < (b)) ? (a) : (b))
#  define MINMAX_DEFINED
#endif
#define EPSDIAG 1.0e-14 /* KKT diagonal min val */
#define EPS 1.0e-6
unsigned MULT = 6;
enum{NO,YES} STEP = NO, prolix = YES;

inline float abs (float v) { return (float)fabs( v); }
inline double abs (double v) { return fabs( v); }
inline long double abs (long double v) { return fabsl( v); }

double dist_func::val(const real_vector& v){
   for(int i=0; i<v.N(); i++) x[i] = v[i];  d = 0.;
   for(int i=0; i<v.N(); i++) d += (x[i]-o[i])*(x[i]-o[i]);
   d = sqrt(d);  return d; }

real_vector& dist_func::grad(real_vector& g){
   g[0] = (x[0]-o[0]) / d;  g[1] = (x[1]-o[1]) / d;
   if(g.N() == 3) g[2] = (x[2]-o[2]) / d;
   return g; }

matrix& dist_func::hessian(matrix& H){ double g[3];
   g[0] = (x[0]-o[0]) / d;  g[1] = (x[1]-o[1]) / d;
   if(H.N() == 3) g[2] = (x[2]-o[2]) / d;
   for(int i=0; i<H.M(); i++){
      H(i+1,i+1) = (1.-g[i]) * (1.+g[i]) / d;
      for(int j=0; j<i; j++)
         H(i+1,j+1) = H(j+1,i+1) = -g[i] * g[j] / d; }
   return H; }

void dist_func::origin(const real_vector& v){
   for(int i=0; i<v.N(); i++) o[i] = v[i];  }

double ray_func::val(const real_vector& v){
   for(int i=0; i<v.N(); i++) x[i] = v[i];  d = e = 0.;
   for(int i=0; i<v.N(); i++) d += (x[i]-o[i])*(x[i]-o[i]);
   for(int i=0; i<v.N(); i++) e += (x[i]-p[i])*(x[i]-p[i]);
   d = sqrt(d);  e = sqrt(e);  return d+e; }

real_vector& ray_func::grad(real_vector& g){
   g[0] = (x[0]-o[0]) / d + (x[0]-p[0]) / e;
   g[1] = (x[1]-o[1]) / d + (x[1]-p[1]) / e;
   if(g.N() == 3) g[2] = (x[2]-o[2]) / d + (x[2]-p[2]) / e;
   return g; }

matrix& ray_func::hessian(matrix& H){ double g[3], h[3];
   g[0] = (x[0]-o[0]) / d;  g[1] = (x[1]-o[1]) / d;
   if(H.N() == 3) g[2] = (x[2]-o[2]) / d;
   h[0] = (x[0]-p[0]) / e;  h[1] = (x[1]-p[1]) / e;
   if(H.N() == 3) h[2] = (x[2]-p[2]) / e;
   for(int i=0; i<H.M(); i++){
      H(i+1,i+1) = (1.-g[i])*(1.+g[i])/d + (1.-h[i])*(1.+h[i])/e;
      for(int j=0; j<i; j++)
         H(i+1,j+1) =
         H(j+1,i+1) = -g[i]*g[j]/d - h[i]*h[j]/e; }
   return H; }

void ray_func::origin(const real_vector& u, const real_vector& v){
   for(int i=0; i<u.N(); i++) o[i] = u[i];
   for(int i=0; i<v.N(); i++) p[i] = v[i];  }

void WaitForIt ()
{
   char ch = '\0';
   fprintf(stderr,"Press ENTER to continue . . . ");
   while (!(ch == '\n' || ch == '\r'))
      ch = (char)getchar();
   return;
}

void spline_fit(double x[], double y[], double g[], unsigned nint)
{
   unsigned n = nint-1;
   sparse T(n, n, n*3-2);
   real_vector d(g+1,n);
   for(int k=1; k<=n-1; k++){
      T(k+1,k) = (x[k+1]-x[k])/6.;            //A(k+1)
      T(k,k+1) = (x[k+1]-x[k])/6.;}           //C(k)
   for(int k=1; k<=n; k++){
      d(k)   = (y[k+1]-y[k])/(x[k+1]-x[k])    //D(k)
             - (y[k]-y[k-1])/(x[k]-x[k-1]);
      T(k,k) = (x[k]-x[k-1]+x[k+1]-x[k])/3.;} //B(k)
   for(int k=2; k<=n; k++){
      T(k,k) -= T(k,k-1)*T(k-1,k)/T(k-1,k-1); //W(k)
      T(k,k-1) /= T(k-1,k-1);   }             //A(k)/W(k-1)
   d.Lsolve( T ).Usolve( T );
}

double splind(double* x, double* y, double* g, unsigned nint, double X)
{
   double d, u, v, w, cku, ckv, Y;
   if(X > x[nint]) return y[nint];
   if(X <= x[0] )  return y[0];
   for(unsigned k=1; k<=nint; k++){
      if(X > x[k]) continue;
      d = x[k] - x[k-1];  u = (X - x[k-1])/d;   v = 1. - u;
      w = d*d/6.;         cku = g[k]*w*u;       ckv = g[k-1]*w*v;
      Y = y[k]*u - cku + y[k-1]*v - ckv + cku*u*u + ckv*v*v;
      break; }
   return Y;
}

void Gram_Schmidt( matrix& e )
{//Korn&Korn p454
   e[0] *= (1./sqrt(e[0]*e[0]));
   for(unsigned i=1; i<e.M(); i++){
      for(unsigned k=0; k<i; k++) e[i] -= (e[k] * (e[i]*e[k]));
      e[i] *= (1./sqrt(e[i]*e[i])); }
}

unsigned steepest_descent(const sparse& A, real_vector& b, double tol)
{//G&VL (10.2.1) p517
   unsigned k = 0; double alf, rsq;
   real_vector x(b.N()), r(b.N());  x.init();  r = b;
   while( (rsq = r*r) > tol ){ k++;
      alf = rsq / (r*(A*r));                         //6
      x += (alf * r);
      r = b - A * x;         }
   b = x;  x.zap();  r.zap();
   return k;
}

double CG_Version_One(const sparse& A, real_vector& b)
{//Beckman (in R&W) p70
   double alf, bet, lam;
   real_vector x(b.N()), r(b.N()), p(b.N()), Ap(b.N());
               x.init();  r = b;  p = b;
   for(unsigned k=0; k<b.N(); k++){
      Ap = A * p;  lam = p * Ap;  alf = (r * p) / lam;
      x += alf * p;
      r -= alf * Ap;
      bet = -(r * Ap) / lam;
      p = r + bet * p;
   }
   b = x;  x.zap();  r.zap();  p.zap();  Ap.zap();
   return r*r;
}

void Pos_Def_Con_Grad(const sparse& A, real_vector& b)
{//G&VL 10.2.7 p524
   unsigned n = A.N();
   if(A.M() != n || b.N() != n) exit(1);
   double* space = new double [3*n];
   memset(space, 0, 3*n*sizeof(double));
   real_vector r(space,n), p(space+n,n), Ap(space+n*2,n);
   double alf, bet, rho_1, rho_2, eps = 1.e-10;

   r = b; b.init(); rho_2 = r*r; eps = (eps*eps*rho_2);
   p = r; Ap = A*p;  alf = rho_2 / (p*Ap);
   b += (p*alf);   r -= (Ap*alf);  rho_1 = r*r;
   r.dumpall();

   for (unsigned j=2; j<=n+3; j++)
   {
      if(rho_1 < eps) break;
      bet = rho_1 / rho_2;  p *= bet;  p += r;
      Ap = A*p; alf = rho_1 / (p*Ap);
      b += (p*alf);   r -= (Ap*alf);  rho_2 = rho_1;  rho_1 = r*r;
      printf("bet %G alf %G ", bet, alf);
      r.dumpall();
   }
   delete space;
}

void Sym_Indef_Bicon_Grad(const sparse& A, real_vector& b)
{
   unsigned n = A.N();  if(A.M() != n || b.N() != n) exit(1);
   double* space = new double [5*n];
   memset(space, 0, 5*n*sizeof(double));
   real_vector r(space,n), s(space+n,n), p(space+n*2,n),
               q(space+n*3,n), Ap(space+n*4,n);
   double alf, bet, rho_1, rho_2, eps = 1.e-10;

   eps = eps*eps*(b*b);
   r = b; b.init(); s = r*A; rho_2 = s*r;
   p = r; q = A*p;  alf = rho_2 / (q*q);
   b += (p*alf);  r -= (q*alf);  s -= ((q*A)*alf);  rho_1 = s*r;

      r.dumpall();
   for (unsigned k=2; k<=n+3; k++)
   {
      if(r*r < eps) break;
      bet = rho_1 / rho_2;  p *= bet;  p += r;  q *= bet;  q += s;
      Ap = A*p; alf = rho_1 / (q*Ap);
      b += (p*alf);   r -= (Ap*alf);  s -= ((q*A)*alf);
      rho_2 = rho_1;  rho_1 = s*r;
      printf("bet %G alf %G\n", bet, alf);
      r.dumpall();
   }
   delete space;
}

double Power_Method(const matrix& A, real_vector& x)
{//After G&VL p351 (7.3.3)
   double l, eps=1.e-12, o = 0., s = 1.;
   real_vector q(x.N()); q = x;

   for(int i=0; i<4*(x.N()); i++){
     l = sqrt(q*q); if(l < eps) break;
     x = (q *= (s/l));
     l = x * (q = A * x);
//   printf("%d %G\n", i, l);  x.dumpall();  WaitForIt();
     if(abs(l-o) < eps) break;  o = l;
     s = (l > 0. ? 1. : -1.);
   }
   q.zap();
   return l;
}

unsigned LanczosTridiag (const sparse& A, real_vector& diag,
//unsigned LanczosTridiag( const matrix& A, real_vector& diag,
                                          real_vector& sub1 )
{//after G&VL p485 alg. 9.2.1
   unsigned j, ncols = A.N();
   if(A.M() != ncols) exit(1);  // also symmetric
   double* space = new double [3*ncols];
   memset (space, 0, 3*ncols*sizeof(double)); diag.init(); sub1.init();
   real_vector v(space,ncols), w(space+ncols,ncols);  w(2) = 1.;
   real_vector t(space+ncols*2,ncols); //w(320) = 1.; //w(1) = 1.;
   double alfa, beta, eps = 1.e-12;    //w(400) = 1.; //w(321) = -1.;
   v += A * w;
   diag(1) = alfa = w * v;

   for(j=2; j<=ncols; j++)
   {
      v -= w * alfa;
      sub1(j) = beta = sqrt( v * v);
      if(beta < eps) break;
      t = v;  v = w * (-beta);  w = t * (1./beta);
      v += A * w;;
      diag(j) = alfa = w * v;
   }
   delete space;
   return j-1;
}

void givens( real_vector& diag, real_vector& sub1, unsigned n,
                                unsigned k, double a, double b )
{//Built for speed. diag(k) = T(k,k), sub1(k) = T(k,k-1).
   double c, s, r, t, u, v;
   if(b == 0.) return;
     else{
       if(fabs(b) > fabs(a)){
         r = -a/b; s = 1./sqrt(1.+r*r); c = s*r;}
         else { r = -b/a; c = 1./sqrt(1.+r*r); s = c*r;}
     }
   unsigned j=k+1, l=k+2;
   if(k>1){
   r = sub1(k); t = sub1(1); sub1(k) = c*r-s*t; }
   r = diag(k); v = sub1(j); diag(k) = c*r-s*v; sub1(j) = s*r+c*v;
   t = diag(j); u = c*v-s*t;                    diag(j) = s*v+c*t;
   r = diag(k);                                 diag(k) = c*r-s*u;
   r = sub1(j); t = diag(j); sub1(j) = c*r-s*t; diag(j) = s*r+c*t;
   if(n>j){ sub1(1) = -s*sub1(l); sub1(l) *= c;}
   return;//sub1(1) is the fugitive term
}

void alg_822( real_vector& diag, real_vector& sub1, unsigned n )
{//diag(k) = T(k,k), sub1(k) = T(k,k-1).
   double d, mu, a, b; unsigned k;
   d = ( diag(n-1) - diag(n) ) * .5;
   a = sub1(n)*sub1(n); mu = sqrt(d*d+a); if(d < 0.) mu = -mu;
   mu = d + mu; mu = diag(n) - a/mu;
   a = diag(1) - mu; b = sub1(2);
   for (k=1; k<n; k++)
   {
      givens( diag, sub1, n, k, a, b );
      if(k > n-2) break;
      a = sub1(k+1); b = sub1(1);
   }                   //sub1(1) is the fugitive term
}

void alg_823( real_vector& diag, real_vector& sub1 )
{//G&VL p424. diag(k) = A(k,k), sub1(k) = A(k,k-1).
   double e=1.0e-12;
   unsigned n=diag.N();
   while ( n > 1 )
   {
      alg_822( diag, sub1, n );
      if( fabs( sub1(n) ) <= ( fabs(diag(n-1)) + fabs(diag(n)) )*e
         && n > 0 )  { sub1(n) = 0.;  n--; }
//    diag.dumpall();  sub1.dumpall();     WaitForIt();
   }
   sub1(1) = 0.;
   return;
}

int INT_PT(const matrix& A, const real_vector& b,
     const real_vector& c, real_vector& x, real_vector& y, double f)
{//Sets up Loop_LP(), for matrix A.
   unsigned m, n, i, j;
   if((m = A.M()) != b.N() || (n = A.N()) != c.N()) exit(1);
   real_vector w(m), z(n);
   int status;
   for(j=0; j<n; j++){ x[j] = 10.0; z[j] = 10.0; }
   for(i=0; i<m; i++){ y[i] = 10.0; w[i] = 10.0; }
   status = Loop_LP( A, b, c, x, z, y, w, f);
   cout << 'w'; w.dumpall();
   cout << 'z'; z.dumpall();
   w.zap(); z.zap();
   cout << "Return status = " << status << endl;
   return status;
}

int INT_PT(const sparse& A, const real_vector& b,
     const real_vector& c, real_vector& x, real_vector& y, double f)
{//Sets up Loop_LP(), for sparse A.
   unsigned m, n, i, j;
   if((m = A.M()) != b.N() || (n = A.N()) != c.N()) exit(1);
   real_vector w(m), z(n);
   int status;
   for(j=0; j<n; j++){ x[j] = 10.0; z[j] = 10.0; }
   for(i=0; i<m; i++){ y[i] = 10.0; w[i] = 10.0; }
   status = Loop_LP( A, b, c, x, z, y, w, f);
   cout << 'w'; w.dumpall();
   cout << 'z'; z.dumpall();
   w.zap(); z.zap();
   cout << "Return status = " << status << endl;
   return status;
}

int Loop_LP(const matrix& A, const real_vector& b,
            const real_vector& c, real_vector& x, real_vector& z,
                  real_vector& y, real_vector& w, double f)
{//Implements RVDB Fig 17.1, for matrix A, using LDLt_QD().
   unsigned i, j, m = A.M(), n = A.N(), nz = m * n;
   real_vector rho(m), sigma(n), u(m+n), dw(m), dz(n);
   real_vector dy(&u[0],m), dx(&u[m],n), E(m), D(n), diag(m+n);
   int status = 0, iter = 0, max_iter = 200;
   double  primal_obj, dual_obj, normr, norms;
   double  gamma, delta = 0.02, mu, theta, r = 0.9;
   while(1)
   {
//  * Compute infeasibilities.
      rho   = -((A*x) -= b); rho -= w;
      sigma = -((y*A) -= c); sigma += z;
      normr = euc( rho );    norms = euc( sigma );
//  * Compute complementarity gap.
      gamma = z * x + y * w;
//  * Compute central path parameter.
      mu = delta * gamma / (double)(n+m);
//  * Print current values.
      if(prolix){
         primal_obj = c * x + f;
         dual_obj   = b * y + f;
         printf("%3d %14.7e %8.1e %14.7e %8.1e %8.1e",
          iter, primal_obj, normr, dual_obj, norms, gamma); }
//  * Check stopping rules.
      if ( normr < EPS && norms < EPS && gamma < EPS ) break;
//  * Set up the diagonal of the reduced KKT matrix.
      for(i=1; i<=m; i++) E(i) = w(i)/y(i);
      for(j=1; j<=n; j++) D(j) = z(j)/x(j);
//    Store -E and D into the Quasidefinite diagonal
      for(i=1; i<=m; i++) // upper left diagonal [-E]
         diag(i)   = -max( E(i), EPSDIAG);
      for(j=1; j<=n; j++) // lower right diagonal [D]
         diag(j+m) =  max( D(j), EPSDIAG);
//  * Define L and decompose
      sparse L(m+n, m+n, nz*MULT);
      L.LDLt_QD( A, diag);
//  * Construct the rhs and backsolve
      for(i=1; i<=m; i++) dy(i) = rho(i)   + w(i) - mu/y(i);
      for(j=1; j<=n; j++) dx(j) = sigma(j) - z(j) + mu/x(j);
      u.Lsolve(L).Dsolve(diag).TLsolve(L);
//  * The rest of the solution
      for(i=1; i<=m; i++) dw(i) = mu/y(i) - w(i) - E(i)*dy(i);
      for(j=1; j<=n; j++) dz(j) = mu/x(j) - z(j) - D(j)*dx(j);

//  * Ratio test to find step length.
      theta = r;
      for(j=1; j<=n; j++) {
         if (theta < -dx(j)/x(j)) { theta = -dx(j)/x(j); }
         if (theta < -dz(j)/z(j)) { theta = -dz(j)/z(j); }
      }
      for(i=1; i<=m; i++) {
         if (theta < -dy(i)/y(i)) { theta = -dy(i)/y(i); }
         if (theta < -dw(i)/w(i)) { theta = -dw(i)/w(i); }
      }
      theta = r/theta;
      if(prolix) printf(" Theta: %8.5f\n", theta);
      if(theta < .00001){ status = -1-iter; break;}

//  * Step to new point
      dx *= theta;  dz *= theta;  dy *= theta;  dw *= theta;
      x  += dx;      z += dz;      y += dy;      w += dw;

      if( prolix && STEP && cin.get() == 'q' ) break;
      if( ++iter > max_iter){ status = iter; break; }
   }// Bottom of Loop
   rho.zap(); sigma.zap(); u.zap(); dw.zap();
   E.zap(); D.zap(); diag.zap(); dz.zap();
   if(prolix) printf(" status: %5d\n", status);
   return status;
}

int Loop_LP(const sparse& A, const real_vector& b,
            const real_vector& c, real_vector& x, real_vector& z,
                  real_vector& y, real_vector& w, double f)
{//Implements RVDB Fig 17.1, for sparse A, using LDLt_QD().
   unsigned i, j, m = A.M(), n = A.N(), nz = A.NZ();
   real_vector rho(m), sigma(n), u(m+n), dw(m), dz(n);
   real_vector dy(&u[0],m), dx(&u[m],n), E(m), D(n), diag(m+n);
   int status = 0, iter = 0, max_iter = 200;
   double  primal_obj, dual_obj, normr, norms;
   double  gamma, delta = 0.02, mu, theta, r = 0.9;
   while(1)
   {
//  * Compute infeasibilities.
      rho   = -((A*x) -= b); rho -= w;
      sigma = -((y*A) -= c); sigma += z;
      normr = euc( rho );    norms = euc( sigma );
//  * Compute complementarity gap.
      gamma = z * x + y * w;
//  * Compute central path parameter.
      mu = delta * gamma / (double)(n+m);
//  * Print current values.
      if(prolix){
         primal_obj = c * x + f;
         dual_obj   = b * y + f;
         printf("%3d %14.7e %8.1e %14.7e %8.1e %8.1e",
          iter, primal_obj, normr, dual_obj, norms, gamma); }
//  * Check stopping rules.
      if ( normr < EPS && norms < EPS && gamma < EPS ) break;
//  * Set up the diagonal of the reduced KKT matrix.
      for(i=1; i<=m; i++) E(i) = w(i)/y(i);
      for(j=1; j<=n; j++) D(j) = z(j)/x(j);
//    Store -E and D into the Quasidefinite diagonal
      for(i=1; i<=m; i++) // upper left diagonal [-E]
         diag(i)   = -max( E(i), EPSDIAG);
      for(j=1; j<=n; j++) // lower right diagonal [D]
         diag(j+m) =  max( D(j), EPSDIAG);
//  * Define L and decompose
      sparse L(m+n, m+n, nz*MULT);
      L.LDLt_QD( A, diag);
//  * Construct the rhs and backsolve
      for(i=1; i<=m; i++) dy(i) = rho(i)   + w(i) - mu/y(i);
      for(j=1; j<=n; j++) dx(j) = sigma(j) - z(j) + mu/x(j);
      u.Lsolve(L).Dsolve(diag).TLsolve(L);
//  * The rest of the solution
      for(i=1; i<=m; i++) dw(i) = mu/y(i) - w(i) - E(i)*dy(i);
      for(j=1; j<=n; j++) dz(j) = mu/x(j) - z(j) - D(j)*dx(j);

//  * Ratio test to find step length.
      theta = r;
      for(j=1; j<=n; j++) {
         if (theta < -dx(j)/x(j)) { theta = -dx(j)/x(j); }
         if (theta < -dz(j)/z(j)) { theta = -dz(j)/z(j); }
      }
      for(i=1; i<=m; i++) {
         if (theta < -dy(i)/y(i)) { theta = -dy(i)/y(i); }
         if (theta < -dw(i)/w(i)) { theta = -dw(i)/w(i); }
      }
      theta = r/theta;
      if(prolix) printf(" Theta: %8.5f\n", theta);
      if(theta < .00001){ status = -1-iter; break;}

//  * Step to new point
      dx *= theta;  dz *= theta;  dy *= theta;  dw *= theta;
      x  += dx;      z += dz;      y += dy;      w += dw;

      if( prolix && STEP && cin.get() == 'q' ) break;
      if( ++iter > max_iter){ status = iter; break; }
   }// Bottom of Loop
   rho.zap(); sigma.zap(); u.zap(); dw.zap();
   E.zap(); D.zap(); diag.zap(); dz.zap();
   if(prolix) printf(" status: %5d\n", status);
   return status;
}

int polygon_vertices(const matrix& A, const real_vector& b, matrix& V)
{//Polygon implied by A*x < b; maximize x(1)*cos(ang) + x(2)*sin(ang)
   unsigned m = b.N(), n = A.N();
   real_vector c(n), x(n), y(m);
   int status, k=0, mv = V.M();
   double angrad=.01, dangrad= -.315, f=0.;  c.init();

   for(int lk=0; lk<20; lk++)
   {
      c(1) = cos(angrad); c(2) = sin(angrad);
      if((status = INT_PT( A, b, c, x, y, f)) != 0 ) return -status;

      if(k == 0 || euc( x - V[k-1] ) > .01)
      {
         V[k] = x; k++;
         cout << "k: " << k << " x"; x.dumpall(); //WaitForIt();
      }
      if(k >= mv) break;
      angrad += dangrad;
   }//bottom lk-loop
   c.zap(); x.zap(); y.zap();
   return k;
}

void INT_PTQ(const matrix& A, const real_vector& b,
      const real_vector& c, real_vector& x, real_vector& y, double f)
{//Sets up RVDB Fig 23.3, for matrix A
   unsigned m, n, i, j;
   if((m = A.M()) != b.N() || (n = A.N()) != c.N()) exit(1);
   real_vector w(m), z(n);
   int status;
   for(j=0; j<n; j++){ x[j] = 10.0; z[j] = 10.0; }
   for(i=0; i<m; i++){ y[i] = 10.0; w[i] = 10.0; }
   status = Loop_QPI( A, b, c, x, z, y, w, f);
   cout << 'w'; w.dumpall();
   cout << 'z'; z.dumpall();
   w.zap(); z.zap();
   cout << "Return status = " << status << endl;
   return;
}

int Loop_QPI (const matrix& A, const real_vector& b,
              const real_vector& c, real_vector& x, real_vector& z,
                    real_vector& y, real_vector& w, double f)
{//RVDB Fig 23.3, for case that Q is the identity, A is matrix.
   unsigned i, j, m = A.M(), n = A.N(), nz = m * n;
   real_vector rho(m), sigma(n), u(m+n), dw(m), dz(n);
   real_vector dx(&u[0],n), dy(&u[n],m), E(m), D(n), diag(m+n);
   int status = 0, iter = 0, max_iter = 200;
   double  primal_obj, dual_obj, normr, norms;
   double  gamma, delta = 0.02, mu, theta, r = 0.9;
   while(1)
   {
//  * Compute infeasibilities.
      rho   = -((A*x) -= b); rho += w;   //sgn w
      sigma = -((y*A) -= c); sigma -= z;  sigma += x; //sign z, +Qx
      normr = euc( rho );    norms = euc( sigma );
//  * Compute complementarity gap.
      gamma = z * x + y * w;
//  * Compute central path parameter.
      mu = delta * gamma / (double)(n+m);
//  * Print current values.
      if(prolix){
         primal_obj = c * x + (x*x)*.5 + f;
         dual_obj   = b * y - (x*x)*.5 + f;
         printf("%3d %14.7e %8.1e %14.7e %8.1e %8.1e",
          iter, primal_obj, normr, dual_obj, norms, gamma); }
//  * Check stopping rules.
      if ( normr < EPS && norms < EPS && gamma < EPS ) break;
//  * Set up the diagonal of the reduced KKT matrix.
      for(i=1; i<=m; i++) E(i) = w(i)/y(i);
      for(j=1; j<=n; j++) D(j) = z(j)/x(j); // need D and E later
//    Store -(D+Q) and E into the Quasidefinite diagonal
      for(i=1; i<=m; i++) // now lower right diagonal [+E]
         diag(i+n) =  max( E(i), EPSDIAG);
      for(j=1; j<=n; j++) // now upper left diagonal [-(D+Q)], Q is I !
         diag(j)   = -max( D(j) + 1., EPSDIAG);
//  * Define L and decompose
      sparse L(m+n, m+n, nz*MULT);
      L.LDLt_QDt( A, diag);
//  * Construct the rhs and backsolve
      for(i=1; i<=m; i++) dy(i) = rho(i)   - w(i) + mu/y(i); //signs
      for(j=1; j<=n; j++) dx(j) = sigma(j) + z(j) - mu/x(j);// incl +Qx
      u.Lsolve(L).Dsolve(diag).TLsolve(L);
//  * The rest of the solution
      for(i=1; i<=m; i++) dw(i) = mu/y(i) - w(i) - E(i)*dy(i);
      for(j=1; j<=n; j++) dz(j) = mu/x(j) - z(j) - D(j)*dx(j);

//  * Ratio test to find step length.
      theta = r;
      for(j=1; j<=n; j++) {
         if (theta < -dx(j)/x(j)) { theta = -dx(j)/x(j); }
         if (theta < -dz(j)/z(j)) { theta = -dz(j)/z(j); }
      }
      for(i=1; i<=m; i++) {
         if (theta < -dy(i)/y(i)) { theta = -dy(i)/y(i); }
         if (theta < -dw(i)/w(i)) { theta = -dw(i)/w(i); }
      }
      theta = r/theta;
      if(prolix) printf(" Theta: %8.5f\n", theta);
      if(theta < .00001){ status = -1-iter; break;}

//  * Step to new point
      dx *= theta;  dz *= theta;  dy *= theta;  dw *= theta;
      x  += dx;      z += dz;      y += dy;      w += dw;

      if( prolix && STEP && cin.get() == 'q' ) break;
      if( ++iter > max_iter){ status = iter; break; }
   }// Bottom of Loop
   rho.zap(); sigma.zap(); u.zap(); dw.zap();
   E.zap(); D.zap(); diag.zap(); dz.zap();
   if(prolix) printf(" status: %5d\n", status);
   return status;
}

void SQP_alg( const matrix& A, const real_vector& b, dist_func& dist,
                                     real_vector& x, real_vector& w)
{//Sets up SQP for dist_func(), linear constraints.
   unsigned m, n, i, j;
   if((m = A.M()) != b.N() || (n = A.N()) != x.N()) exit(1);
   sparse C(m+n, m+n, m + n + m*n + n*(n-1));
   matrix H(n,n);
   real_vector y(m), c(n);
   for(j=0; j<n; j++)  x[j] = 10.0;
   for(i=0; i<m; i++){ y[i] = 10.0; w[i] = 1.0; }
   for(i=1; i<=m; i++)
      for(j=1; j<=n; j++)
         if(fabs(A(i,j)) > EPSDIAG) C.stash(j+m,i) = A(i,j);
 {//Implements SQP for dist_func().
   real_vector rho(m), sigma(n), u(m+n), dw(m);
   real_vector dy(&u[0],m), dx(&u[m],n), E(m), diag(m+n);
   int status = 0, iter = 0, max_iter = 200;
   double  obj_val, normr, norms;
   double  gamma, delta = 0.02, mu, theta, r = 0.9;
   while(1)
   {
//  * Current values
      obj_val = dist.val(x);
      dist.grad(c);
      dist.hessian(H);
//  * Infeasibilities
      rho   = -((A*x) -= b);  rho += w;   //nc
      sigma = -((y*A) -= c);           //del z & +Qx
      normr = euc( rho );     norms = euc( sigma );
//  * Complementarity gap & central path parameter
      gamma = y * w;       // del x*z
      mu = delta * gamma / (double)m; //(n+m);
//  * Print current values.
      if(prolix){
         printf("%3d %14.7e %8.1e %8.1e %8.1e",
          iter, obj_val, normr, norms, gamma ); }
//  * Check stopping rules.
      if ( normr < EPS && norms < EPS && gamma < EPS ) break;
//  * Set up the diagonal of the reduced KKT matrix.
      for(i=1; i<=m; i++) E(i) = w(i)/y(i);
//    Store -(Q) and E into the Quasidefinite diagonal
      for(i=1; i<=m; i++) // upper left diagonal [+E]
         C(i,i)     =  max( E(i), EPSDIAG);
      for(j=1; j<=n; j++){ // lower right diagonal [-(H)]
         C(j+m,j+m) = -max( H(j,j), EPSDIAG);
         for(i=j+1; i<=n; i++) C(i+m,j+m) = -H(i,j); } //off-diag H
//    Define L and decompose
      sparse L(m+n, m+n, C.NZ()*MULT );
      L.LDLt( C, diag);
//    Construct the rhs and backsolve
      for(i=1; i<=m; i++) dy(i) = rho(i) - w(i) + mu/y(i);//f24.1 wrong
      for(j=1; j<=n; j++) dx(j) = sigma(j); // + z(j) - mu/x(j);
      u.Lsolve(L).Dsolve(diag).TLsolve(L);
//    The rest of the solution
      for(i=1; i<=m; i++) dw(i) = mu/y(i) - w(i) - E(i)*dy(i);

//  * Ratio test to find step length
      theta = r;
      for(i=1; i<=m; i++) if(theta < -dw(i)/w(i)) theta = -dw(i)/w(i);
      for(i=1; i<=m; i++) if(theta < -dy(i)/y(i)) theta = -dy(i)/y(i);
      theta = r/theta;
      if(prolix) printf(" Theta: %8.5f\n", theta);
      if(theta < .00001){ status = -1-iter; break; }

//  * Step to new point
      dx *= theta;  dy *= theta;  dw *= theta;
      x  += dx;      y += dy;      w += dw;

      if( prolix && STEP && cin.get() == 'q' ) break;
      if( ++iter > max_iter){ status = iter; break; }
   }// Bottom of Loop
   rho.zap(); sigma.zap(); u.zap(); dw.zap();
   E.zap(); diag.zap();
   if(prolix) printf(" status: %5d\n", status);
 }
   y.zap(); c.zap();
   return;
}

void SQP_alg( const matrix& A, const real_vector& b, ray_func& path,
                                     real_vector& x, real_vector& w)
{//Sets up SQP for ray_func(), linear constraints.
   unsigned m, n, i, j;
   if((m = A.M()) != b.N() || (n = A.N()) != x.N()) exit(1);
   sparse C(m+n, m+n, m + n + m*n + n*(n-1));
   matrix H(n,n);
   real_vector y(m), c(n);
   for(j=0; j<n; j++)  x[j] = 10.0;
   for(i=0; i<m; i++){ y[i] = 10.0; w[i] = 1.0; }
   for(i=1; i<=m; i++)
      for(j=1; j<=n; j++)
         if(fabs(A(i,j)) > EPSDIAG) C.stash(j+m,i) = A(i,j);
 {//Implements SQP for ray_func().
   real_vector rho(m), sigma(n), u(m+n), dw(m);
   real_vector dy(&u[0],m), dx(&u[m],n), E(m), diag(m+n);
   int status = 0, iter = 0, max_iter = 100;
   double  obj_val, normr, norms;
   double  gamma, delta = 0.02, mu, theta, r = 0.9;
   while(1)
   {
//  * Current values
      obj_val = path.val(x);
      path.grad(c);
      path.hessian(H);
//  * Infeasibilities
      rho   = -((A*x) -= b);  rho += w;   //nc
      sigma = -((y*A) -= c);           //del z & +Qx
      normr = euc( rho );     norms = euc( sigma );
//  * Complementarity gap & central path parameter
      gamma = y * w;       // del x*z
      mu = delta * gamma / (double)m; //(n+m);
//  * Print current values.
      if(prolix){
         printf("%3d %14.7e %8.1e %8.1e %8.1e",
          iter, obj_val, normr, norms, gamma ); }
//  * Check stopping rules.
      if ( normr < EPS && norms < EPS && gamma < EPS ) break;
//  * Set up the diagonal of the reduced KKT matrix.
      for(i=1; i<=m; i++) E(i) = w(i)/y(i);
//    Store -(Q) and E into the Quasidefinite diagonal
      for(i=1; i<=m; i++) // upper left diagonal [+E]
         C(i,i)     =  max( E(i), EPSDIAG);
      for(j=1; j<=n; j++){ // lower right diagonal [-(H)]
         C(j+m,j+m) = -max( H(j,j), EPSDIAG);
         for(i=j+1; i<=n; i++) C(i+m,j+m) = -H(i,j); } //off-diag H
//    Define L and decompose
      sparse L(m+n, m+n, C.NZ()*MULT );
      L.LDLt( C, diag);
//    Construct the rhs and backsolve
      for(i=1; i<=m; i++) dy(i) = rho(i) - w(i) + mu/y(i);//f24.1 wrong
      for(j=1; j<=n; j++) dx(j) = sigma(j); // + z(j) - mu/x(j);
      u.Lsolve(L).Dsolve(diag).TLsolve(L);
//    The rest of the solution
      for(i=1; i<=m; i++) dw(i) = mu/y(i) - w(i) - E(i)*dy(i);

//  * Ratio test to find step length
      theta = r;
      for(i=1; i<=m; i++) if(theta < -dw(i)/w(i)) theta = -dw(i)/w(i);
      for(i=1; i<=m; i++) if(theta < -dy(i)/y(i)) theta = -dy(i)/y(i);
      theta = r/theta;
      if(prolix) printf(" Theta: %8.5f\n", theta);
      if(theta < .00001){ status = -1-iter; break; }

//  * Step to new point
      dx *= theta;  dy *= theta;  dw *= theta;
      x  += dx;      y += dy;      w += dw;

      if( prolix && STEP && cin.get() == 'q' ) break;
      if( ++iter > max_iter){ status = iter; break; }
   }// Bottom of Loop
   rho.zap(); sigma.zap(); u.zap(); dw.zap();
   E.zap(); diag.zap();
   if(prolix) printf(" status: %5d\n", status);
 }
   y.zap(); c.zap();
   return;
}
