/* method.h                                                     12/28/00

     Copyright (c) John R. Berryhill 2000.  All Rights Reserved.

*/
#ifndef METHOD_H_INCLUDED
#define METHOD_H_INCLUDED
#include <stdio.h>
#include <math.h>
class matrix;
class sparse;
class real_vector;
class pivot;

struct dist_func{
   double d, o[3], x[3];
   dist_func(double=0., double=0., double=0.);
   double val(const real_vector&);
   real_vector& grad(real_vector&);
   matrix& hessian(matrix&);
   void origin(const real_vector&);
};

inline dist_func::dist_func(double x, double y, double z){
   o[0] = x; o[1] = y; o[2] = z; }

struct ray_func{
   double d, e, o[3], p[3], x[3];
   ray_func();
   double val(const real_vector&);
   real_vector& grad(real_vector&);
   matrix& hessian(matrix&);
   void origin(const real_vector&, const real_vector&);
};

inline ray_func::ray_func(){
   o[0] = o[1] = o[2] = p[0] = p[1] = p[2] = 0.; }

void WaitForIt ();
void spline_fit (double*, double*, double*, unsigned);
double splind (double*, double*, double*, unsigned, double);

void Gram_Schmidt (matrix&);
unsigned steepest_descent (const sparse&, real_vector&, double);
double CG_Version_One (const sparse&, real_vector&);
void Pos_Def_Con_Grad (const sparse&, real_vector&);
void Sym_Indef_Bicon_Grad (const sparse&, real_vector&);

double Power_Method (const matrix&, real_vector&);
unsigned LanczosTridiag (const sparse&, real_vector&, real_vector&);
unsigned LanczosTridiag (const matrix&, real_vector&, real_vector&);
void givens (real_vector&, real_vector&,
             unsigned, unsigned, double, double);
void alg_822 (real_vector&, real_vector&, unsigned);
void alg_823 (real_vector&, real_vector&);

int INT_PT (const matrix&, const real_vector&, const real_vector&,
                                 real_vector&, real_vector&, double);
int INT_PT (const sparse&, const real_vector&, const real_vector&,
                                 real_vector&, real_vector&, double);
int Loop_LP (const matrix&, const real_vector&, const real_vector&,
        real_vector&, real_vector&, real_vector&, real_vector&, double);
int Loop_LP (const sparse&, const real_vector&, const real_vector&,
        real_vector&, real_vector&, real_vector&, real_vector&, double);
int polygon_vertices(const matrix&, const real_vector&, matrix&);
void INT_PTQ (const matrix&, const real_vector&, const real_vector&,
                                   real_vector&, real_vector&, double);
int Loop_QPI (const matrix& A, const real_vector& b, const real_vector&,
        real_vector&, real_vector&, real_vector&, real_vector&, double);
void SQP_alg (const matrix&, const real_vector&, dist_func&,
                                   real_vector&, real_vector&);
void SQP_alg (const matrix&, const real_vector&, ray_func&,
                                   real_vector&, real_vector&);

#endif METHOD_H

