/* quatern.h                                                    12/28/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/

#include <complex.h>

#if !defined( QUATERN_H )
#define QUATERN_H

class quatern
{
private:

   complex re, im;

public:

   quatern(complex, complex);
   quatern(double, double, double, double);
   quatern(const quatern&);
   quatern();
   ~quatern();

   quatern& operator=(const quatern&);
   quatern& operator+=(quatern&);
   quatern& operator-=(quatern&);
   quatern& operator*=(quatern&);
   quatern& operator*=(double);
   quatern& operator/=(double);

   friend double real(quatern&);
   friend double imag(quatern&);
   friend double jmag(quatern&);
   friend double kmag(quatern&);
   friend double norm(quatern&);   // (squared)

   friend quatern operator+(quatern&, quatern&);
   friend quatern operator-(quatern&, quatern&);
   friend quatern operator*(quatern&, quatern&);
   friend quatern operator*(quatern&, double);
   friend int operator==(quatern&, quatern&);
   friend int operator!=(quatern&, quatern&);
};
   quatern conj(quatern&);
   quatern operator/(quatern&, double);
   quatern inv(const quatern&);
   double arg(const quatern&);    // the angle in the plane

inline quatern::quatern( complex x, complex im_val)
{
    re = x;  im = im_val;
}

inline quatern::quatern( double a, double b, double c, double d)
{
    re = complex (a, b);  im = complex (c, d);
}

/*inline*/ quatern::quatern(const quatern& q)
{
    re = q.re;  im = q.im;
}

inline quatern::quatern()
{
    re = im = complex(0.);
}

inline quatern::~quatern()
{
}

inline quatern& quatern::operator=( const quatern& q2)
{
    re = q2.re;  im = q2.im;  return *this;
}

inline quatern& quatern::operator+=( quatern& q2)            //$
{
    re += q2.re;  im += q2.im;  return *this;
}

inline quatern& quatern::operator-=( quatern& q2)            //$
{
    re -= q2.re;  im -= q2.im;  return *this;
}

inline quatern& quatern::operator*=( double x)
{
    re *= x;  im *= x;  return *this;
}

inline quatern& quatern::operator/=( double x)
{
    re /= x;  im /= x;  return *this;
}

inline double real( quatern& q)                              //$
{
    return real( q.re );
}

inline double imag( quatern& q)                              //$
{
    return imag( q.re );
}

inline double jmag( quatern& q)                              //$
{
    return real( q.im );
}

inline double kmag( quatern& q)                              //$
{
    return imag( q.im );
}

/*inline*/ quatern conj( quatern& q)                         //$
{
    return quatern( real(q), -imag(q), -jmag(q), -kmag(q));
}

inline double norm( quatern& q)                              //$
{
    return norm( q.re ) + norm( q.im );
}

inline double arg( quatern& q)                               //$
{
    return acos( real(q) / sqrt( norm(q) ) );
}

/*inline*/ quatern inv( quatern& q)                          //$
{
    return conj(q) / norm(q);
}

inline quatern operator+( quatern& q1, quatern& q2)          //$
{
    return quatern( q1.re + q2.re, q1.im + q2.im);
}

inline quatern operator-( quatern& q1, quatern& q2)          //$
{
    return quatern( q1.re - q2.re, q1.im - q2.im);
}

inline quatern operator*( quatern& q1, quatern& q2)          //$
{
    return quatern( q1.re*q2.re - q1.im*conj(q2.im),
                    q1.re*q2.im + q1.im*conj(q2.re));
}

inline quatern operator*( quatern& q, double x)              //$
{
    return quatern( q.re*x, q.im*x);
}

/*inline*/ quatern operator/( quatern& q, double x)          //$
{
    return quatern( real(q)/x, imag(q)/x, jmag(q)/x, kmag(q)/x);
}

inline int operator==( quatern& q1, quatern& q2)             //$
{
    return q1.re == q2.re && q1.im == q2.im;
}

inline int operator!=( quatern& q1, quatern& q2)             //$
{
    return q1.re != q2.re || q1.im != q2.im;
}
  //$ means prefer "const quatern&" in argument list
  //  if your complex implementation allows

#include <iostream.h>

ostream& operator<<(ostream& s, quatern& q)
{
    s << "(" << real(q) << ", " << imag(q) << ", "
             << jmag(q) << ", " << kmag(q) << ")\n"; return s;
}

istream& operator>>(istream&, quatern&);

#endif  // QUATERN_H

