/* sparse.cpp                                                   12/28/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/
#include "C:\jrb32\book\source\sparse.h"
#include "C:\jrb32\book\source\vector.h"
#include "C:\jrb32\book\source\matrix.h"
#include <stdio.h>
#include <stdlib.h>

enum{NO,YES} DEBUG = NO;

pivot::pivot(unsigned i, unsigned j)
{
   m = i; n = j; R = new unsigned [m+n]; C = R + m;
   unsigned k;
   for (k=1; k<=m; k++) R[k-1] = k;
   for (k=1; k<=n; k++) C[k-1] = k;
}

void pivot::dumpall (unsigned l) const
{
   unsigned i=m, j=n, k;
   if  (l > 0) i = j = l;
   for (k=0; k<i; k++) cout << "|" << R[k];
   cout << "[R]\n";
   for (k=0; k<j; k++) cout << "|" << C[k];
   cout << "[C]\n";
}

void pivot::init()
{
   unsigned k;
   for (k=1; k<=m; k++) R[k-1] = k;
   for (k=1; k<=n; k++) C[k-1] = k;
}

void pivot::swaprows (const unsigned i, const unsigned j)
{
   unsigned k;
   k = R[j-1]; R[j-1] = R[i-1]; R[i-1] = k;
}

void pivot::swapcols (const unsigned i, const unsigned j)
{
   unsigned k;
   k = C[j-1]; C[j-1] = C[i-1]; C[i-1] = k;
}

sparse::sparse(const unsigned i, const unsigned j, const unsigned max)
{
   mrows = i;  ncols = j;  kount = mrows+ncols;  size = max+kount;
   row = new element* [mrows];  col = new element* [ncols];
   e = new element[size];  checkptr(e);
   memset(e, 0, size*sizeof(element));  unsigned k;
   for (k=0; k<mrows; k++) row[k] = e + k;
   for (k=0; k<ncols; k++) col[k] = e + mrows + k;
}

sparse::~sparse()
{                                     if (DEBUG) printf("del %Fp \n",e);
   delete e;  delete row;  delete col;
}

double& sparse::operator () (const unsigned i, const unsigned j)
{
   element* elem = row[i-1];  unsigned k;

   while ((k = elem->next) != 0){ elem = e+k;
      if(elem->col == j) return elem->value; }

   if(kount >= size){ dumpall(kount-mrows-ncols-10); exit(1); }
   k = kount; e[k].row = i; e[k].col = j;
   elem = row[i-1];                       // insert e[k] at row head
   elem->col++;                           // update rowcount
   e[k].next = elem->next; elem->next = kount++;

   elem = col[j-1];  elem->row++;    // e[k] col head; update colcount
   e[k].mext = elem->mext; elem->mext = k;

   return e[k].value;
}

double& sparse::stash (const unsigned i, const unsigned j)
{//Use when guaranteed that (i,j) does not already exist
   element* elem = row[i-1];  unsigned k = kount;
   if(kount >= size){ dumpall(kount-mrows-ncols-10); exit(1); }
   e[k].row = i; e[k].col = j;

   elem->col++;                           // update rowcount
   e[k].next = elem->next; elem->next = kount++;

   elem = col[j-1];  elem->row++;    // e[k] col head; update colcount
   e[k].mext = elem->mext; elem->mext = k;

   return e[k].value;
}

double sparse::getval (const unsigned i, const unsigned j) const
{
   element* elem = row[i-1];  unsigned k;
   while ((k = elem->next) != 0){ elem = e+k;
        if(elem->col == j) return elem->value; }
   elem = col[j-1];                              //try both ways
   while ((k = elem->mext) != 0){ elem = e+k;
        if(elem->row == i) return elem->value; }
   return 0.;
}

void sparse::unlink (const unsigned i, const unsigned j)
{//Isolate a specific element
   element* elem = row[i-1];  unsigned k;
   while ((k = elem->next) != 0){
      if(e[k].col == j){
         elem->next = e[k].next;  e[k].next = e[k].col = 0;
         rowcount(i)--;   break; }
      elem = e+k;}

   elem = col[j-1];
   while ((k = elem->mext) != 0){
      if(e[k].row == i){
         elem->mext = e[k].mext;  e[k].mext = e[k].row = 0;
         colcount(j)--;   break; }
      elem = e+k;}
}

void sparse::row_unlink( const unsigned i, const unsigned j)
{  element* seek;  unsigned k, r;
   seek = row[i-1];  r = seek->row;  //original row number
   while( (k = seek->next) != 0){ if( e[k].col == j){
      seek->next = e[k].next;  e[k].next = 0;
      e[k].row = r;  rowcount(i)--;  break; } seek = e+k; }
}

void sparse::col_unlink( const unsigned i, const unsigned j)
{  element* seek;  unsigned k, c;
   seek = col[j-1];  c = seek->col;  //original col number
   while( (k = seek->mext) != 0){ if( e[k].row == i){
      seek->mext = e[k].mext;  e[k].mext = 0;
      e[k].col = c;  colcount(j)--;  break; } seek = e+k; }
}

void sparse::reset (const pivot& P)
{//Set the L\U rows and cols to final values
   element *elem; unsigned i, j, k;
   if(mrows != P.m || ncols != P.n) exit(1);
   for(i=0; i<mrows; i++){ elem = row[i];
      while ((k = elem->next) != 0){ elem = e+k;
         j = elem->col;  elem->col = P.C[j-1]; }
   }
   for(j=0; j<ncols-1; j++){ elem = col[j];
      while ((k = elem->mext) != 0){ elem = e+k;
         i = elem->row;  elem->row = P.R[i-1]; }
   }
}

double sparse::row_maxab (const unsigned i)
{
   element* elem = row[i-1];  unsigned k;  double maxab = 0.;
   while( (k = elem->next) != 0){ elem = e+k;
      if(fabs(elem->value) > maxab) maxab = fabs(elem->value); }
   return maxab;
}

void sparse::row_scale (const unsigned i, const double maxab)
{
   element* elem = row[i-1];  unsigned k;
   while( (k = elem->next) != 0){ elem = e+k;  elem->value *= maxab; }
}

void sparse::rowneg (const unsigned i)
{//negate each nonzero in row
   element* elem = row[i-1];  unsigned k;
   while ((k = elem->next) != 0){ elem = e+k;
         (elem->value) *= -1.; }
}

unsigned sparse::inspect () const
{
   unsigned k = size - mrows - ncols;
   cout << "Sparse matrix " << mrows << " rows by " << ncols
        << " columns with " << k     << " max nonzeros\n";
   cout << "       First: " << (mrows+ncols) << e[mrows+ncols] << "\n";
   cout << "       Last:  " << (kount-1) << e[kount-1];
   k = size - kount;
   cout << " Remaining: " << k << "\n";
   print_stats( *this );
   return k;
}

void sparse::dumprow (const unsigned i) const
{
   element* elem = row[i-1];  unsigned k;
   while ((k = elem->next) != 0){ elem = e+k;
      cout << k << *elem << "\n"; }
}

void sparse::dumpcol (const unsigned j) const
{
   element* elem = col[j-1];  unsigned k;
   while ((k = elem->mext) != 0){ elem = e+k;
      cout << k << *elem << "\n"; }
}

void sparse::dumpall (const unsigned skip) const
{
   for (unsigned k=mrows+ncols+skip; k<kount; k++)
       cout << k << e[k] << "\n";
   cout << "\n";
}

void sparse::swaprows (const unsigned i, const unsigned j)
{
   element* elem = row[i-1];  row[i-1] = row[j-1];  row[j-1] = elem;
   unsigned k;
   while ((k = elem->next) != 0){ elem = e+k; elem->row = j; }
   elem = row[i-1];
   while ((k = elem->next) != 0){ elem = e+k; elem->row = i; }
}

void sparse::swapcols (const unsigned i, const unsigned j)
{
   element* elem = col[i-1];  col[i-1] = col[j-1];  col[j-1] = elem;
   unsigned k;
   while ((k = elem->mext) != 0){ elem = e+k; elem->col = j; }
   elem = col[i-1];
   while ((k = elem->mext) != 0){ elem = e+k; elem->col = i; }
}

unsigned sparse::rowpop (const unsigned i, const unsigned j) const
{//this actually counts
   element* elem = row[i-1];  unsigned k, count=0;

   while ((k = elem->next) != 0){ elem = e+k;
      if (elem->col >= j) count++; }
   return count;
}

unsigned sparse::colpop (const unsigned i, const unsigned j) const
{//this actually counts
   element* elem = col[j-1];  unsigned k, count=0;

   while ((k = elem->mext) != 0){ elem = e+k;
      if (elem->row >= i) count++; }
   return count;
}

void sparse::LDLt (const sparse& A, real_vector& d)
{//Doolittle elimination output to lower-triangular *this
   unsigned i,j,k,l,r;  double lij, lij_dj, diagi;
   element *elem, *prev;
   real_vector tmp (ncols);  tmp.init();
   for (i=1; i<=d.N(); i++) d(i) = A.getval(i,i);

   lij_dj = A.getval(2,1);
   if(lij_dj != 0.){
      lij = lij_dj / d(1);
      d(2) -= lij * lij_dj;
      stash(2,1) = lij; }

   for (i=3; i<=mrows; i++)  //row-by-row computation of L
   {
      diagi = d(i);

//    Dump row i of A into tmp, checking j < i:
      elem = A.row[i-1];
      while ((k = elem->next) != 0){ elem = A.e+k;
         if((j = elem->col) >= i) continue;
         tmp(j) = elem->value;  }

//    Process row i from left to right:
      for (j=1; j<i; j++)
      {
         if (tmp(j) == 0.) continue;
         if ( d(j)  == 0.) continue;//( redundant,theoretically)
         lij_dj = tmp(j); lij = lij_dj / d(j);  tmp(j) = 0.;
         diagi -= lij * lij_dj;

//       Combine lij_dj and previous lij from this column
         prev = col[j-1];
         while ((l = prev->mext) != 0){ prev = e+l;
            r = prev->row;  tmp(r) -= lij_dj * (prev->value);
         }
         stash(i,j) = lij;
      }
      d(i) = diagi; // This row is now final
   }
   tmp.zap();
   return;
}

void sparse::LDLt_QD (const sparse& A, real_vector& d)
{//Input A(m,n): A(j,i-m) -> L(i,j): L(m+n,m+n)
   unsigned i,j,k,l,m,r;  double lij, lij_dj, diagi;
   element *elem, *prev;
   real_vector tmp (mrows);  tmp.init();
   if(((m = A.M()) + A.N()) != mrows) exit(1);

 //First m rows are no-op!
   for (i=m+1; i<=mrows; i++)  //row-by-row computation of L
   {
      diagi = d(i);
      elem = A.col[i-m-1];         // dump A-col into tmp
      while((k = elem->mext) != 0){ elem = A.e+k;
         if((j = elem->row)  >= i) continue; //never
         tmp(j) = elem->value; }

      for (j=1; j<i; j++)           // do cols in ascending order
      {
         if (tmp(j) == 0.) continue;
         if ( d(j)  == 0.) continue;//(redundant,theoretically)
         lij_dj = tmp(j); lij = lij_dj / d(j);  tmp(j) = 0.;
         diagi -= lij * lij_dj;
         prev = col[j-1];           // col j of L so far
         while ((l = prev->mext) != 0)
         {
            prev = e+l;  r = prev->row;
            tmp(r) -= lij_dj * (prev->value);
         }
         stash(i,j) = lij;
      }
      d(i) = diagi;
   }
   tmp.zap();
   return;
}

void sparse::LDLt_QDt (const sparse& A, real_vector& d)
{//Input A(m,n): A(i-n,j) -> L(i,j): L(m+n,m+n)
   unsigned i,j,k,l,n,r;  double lij, lij_dj, diagi;
   element *elem, *prev;
   real_vector tmp (mrows);  tmp.init();
   if(((n = A.N()) + A.M()) != mrows) exit(1);

 //First n rows are no-op!
   for (i=n+1; i<=mrows; i++)  //row-by-row computation of L
   {
      diagi = d(i);
      elem = A.row[i-n-1];         // dump A-row into tmp
      while((k = elem->next) != 0){ elem = A.e+k;
         if((j = elem->col)  >= i) continue; //never
         tmp(j) = elem->value; }

      for (j=1; j<i; j++)           // do cols in ascending order
      {
         if (tmp(j) == 0.) continue;
         if ( d(j)  == 0.) continue;//(redundant,theoretically)
         lij_dj = tmp(j); lij = lij_dj / d(j);  tmp(j) = 0.;
         diagi -= lij * lij_dj;
         prev = col[j-1];           // col j of L so far
         while ((l = prev->mext) != 0)
         {
            prev = e+l;  r = prev->row;
            tmp(r) -= lij_dj * (prev->value);
         }
         stash(i,j) = lij;
      }
      d(i) = diagi;
   }
   tmp.zap();
   return;
}

void sparse::LDLt_QD (const matrix& A, real_vector& d)
{//Input A(m,n): A(j,i-m) -> L(i,j): L(m+n,m+n)
   element *prev;  unsigned i,j,l,m,r;  double lij, lij_dj, diagi;
   real_vector tmp (mrows);  tmp.init();
   if(((m = A.M()) + A.N()) != mrows) exit(1);

 //First m rows are no-op!
   for (i=m+1; i<=mrows; i++)  //row-by-row computation of L
   {
      diagi = d(i);
      for (j=1; j<=m; j++) tmp(j) = A(j,i-m); // dump A-col into tmp

      for (j=1; j<i; j++)           // do cols in ascending order
      {
         if (tmp(j) == 0.) continue;
         if ( d(j)  == 0.) continue;//(redundant,theoretically)
         lij_dj = tmp(j); lij = lij_dj / d(j);  tmp(j) = 0.;
         diagi -= lij * lij_dj;
         prev = col[j-1];           // col j of L so far
         while ((l = prev->mext) != 0)
         {
            prev = e+l;  r = prev->row;
            tmp(r) -= lij_dj * (prev->value);
         }
         stash(i,j) = lij;
      }
      d(i) = diagi;
   }
   tmp.zap();
   return;
}

void sparse::LDLt_QDt (const matrix& A, real_vector& d)
{//Input A(m,n): A(i-n,j) -> L(i,j): L(m+n,m+n)
   element *prev;  unsigned i,j,l,n,r;  double lij, lij_dj, diagi;
   real_vector tmp (mrows);  tmp.init();
   if(((n = A.N()) + A.M()) != mrows) exit(1);

 //First n rows are no-op!
   for (i=n+1; i<=mrows; i++)  //row-by-row computation of L
   {
      diagi = d(i);
      for (j=1; j<=n; j++) tmp(j) = A(i-n,j); // dump A-row into tmp

      for (j=1; j<i; j++)           // do cols in ascending order
      {
         if (tmp(j) == 0.) continue;
         if ( d(j)  == 0.) continue;//(redundant,theoretically)
         lij_dj = tmp(j); lij = lij_dj / d(j);  tmp(j) = 0.;
         diagi -= lij * lij_dj;
         prev = col[j-1];           // col j of L so far
         while ((l = prev->mext) != 0)
         {
            prev = e+l;  r = prev->row;
            tmp(r) -= lij_dj * (prev->value);
         }
         stash(i,j) = lij;
      }
      d(i) = diagi;
   }
   tmp.zap();
   return;
}

unsigned sparse::Crout_LU( pivot& P )
{//Crout elimination column by column
   unsigned i,j,k,l,p;  double Ujj, Ukj;
   element *elem;
   real_vector tmp(mrows);  tmp.init();

   for (j=1; j<=ncols; j++)  //col-by-col computation of L\U
   {
//    Dump col j into tmp:
      elem = col[j-1];
      while((l = elem->mext) != 0){ elem = e+l;
             i = elem->row;  tmp(i) = elem->value; }

//    Recursion on the first j-1 entries of tmp
      for(k=1; k<j; k++){
         if((Ukj = tmp(k)) == 0.) continue;
//       prior column k of L:
         elem = col[k-1];
         while((l = elem->mext) != 0){ elem = e+l;
            if((i = elem->row) <= k) continue;      //since it's Uik
            tmp(i) -= (elem->value)*Ukj;   }
      }
//    Finalize col j from top to j-1:
      for(i=1; i<j; i++){
         if( tmp(i) == 0. && getval(i,j) == 0.) continue;
         (*this)(i,j) = tmp(i); tmp(i) = 0.;          //Uij
      }
//    Find the pivot
      Ujj = tmp(j);  p = j;
      for(i=j+1; i<=mrows; i++){
         if(fabs(tmp(i)) <= fabs(Ujj)) continue;
         Ujj = tmp(i);  p = i;
      }
      if( Ujj == 0.) break;  //return j-1;
//    Swap rows
      if(p != j){
         swaprows(j,p);
         Ukj = tmp(j);  tmp(j) = tmp(p);  tmp(p) = Ukj;
         P.swaprows(j,p);
//         k = P.R[j-1]; P.R[j-1] = P.R[p-1]; P.R[p-1] = k;
      }
//    Finalize col j from j to bottom:
      (*this)(j,j) = Ujj;  tmp(j) = 0.;
      for(i=j+1; i<=mrows; i++){
         if( tmp(i) == 0. && getval(i,j) == 0.) continue;
         (*this)(i,j) = tmp(i)/Ujj;  tmp(i) = 0.;   //Lij
      }
   }
   tmp.zap();
   return j-1;
}

unsigned sparse::Gauss_LU()
{
   element* elem;  unsigned i,j,k,l,r,rank;
   element* cand;  double Lir,Urr;

   for( r=1; r<mrows; r++)
   {
      if( (Urr = (*this)(r,r)) == 0.) break;

      elem = col[r-1];
      while( (k = elem->mext) != 0){ elem = e+k;
         if( (i = elem->row ) <= r) continue;
         Lir = (elem->value) /= Urr;

         cand = row[r-1];
         while( (l = cand->next) != 0){ cand = e+l;
            if( (j = cand->col)  <= r) continue;
            (*this)(i,j) -= Lir * cand->value;     //Urj
      }  }  //bottom outer-product update
   }//bottom r-loop
   if(r == mrows && getval(r,r) != 0.) rank = mrows;
   else rank = r-1;
   return rank;
}

unsigned sparse::Gauss_LU( pivot& P )
{//
   element* elem;  unsigned i,j,k,l,r,rank,*idx,n;
   double Lir,Urr;
   real_vector Urj(ncols);   idx = P.R;

   for (i=0; i<mrows; i++) (row[i])->row = P.R[i];
   for (j=0; j<ncols; j++) (col[j])->col = P.C[j];

   for (r=1; r<mrows; r++)
   {
      if ((Urr = Pivot( r ) ) == 0.) break;

      elem = row[r-1];  n = 0;  Urr = 0.;        //copy pivot row
      while( (k = elem->next) != 0){ elem = e+k;
         if( (j = elem->col)  <  r) continue;    //includes diag
         Urj[n] = elem->value;  idx[n] = j;  n++;
           if(j == r){ n--;  Urr = Urj[n]; }     //the diagonal
         col_unlink(r,j);
      }                                   //Row r of U now complete

      if( Urr == 0.) break;

      elem = col[r-1];                           //work pivot column
      while( (k = elem->mext) != 0){ elem = e+k;
         if( (i = elem->row ) <= r) continue;
         Lir = (elem->value) /= Urr;
         row_unlink(i,r);

         for(l=0; l<n; l++){                           //outer product
            j = idx[l];  (*this)(i,j) -= Lir * Urj[l]; }
      }                                   //Col r of L is now complete
   }             //Bottom of r-loop
   col_unlink(r,r);

   for (i=0; i<mrows; i++) {k = (row[i])->row; P.R[k-1] = i + 1;}
   for (j=0; j<ncols; j++) {k = (col[j])->col; P.C[k-1] = j + 1;}
   reset(P); //update row and col numbers

   for (i=0; i<mrows; i++) {k = (row[i])->row; P.R[i] = k;}
   for (j=0; j<ncols; j++) {k = (col[j])->col; P.C[j] = k;}
   Urj.zap();
   if(r == mrows && getval(r,r) != 0.) rank = mrows;
   else rank = r-1;
   return rank;
}

#define THRESHOLD 0.1
double sparse::Pivot( const unsigned r)
{
   element* elem;  unsigned i,k,p;
   unsigned count, cost;  double v,val,Urr=0.;  p = r;

// Ensure good starting column
   while ((count = colcount(p)) == 0 && p < ncols) p++;
   if ((cost = count) == 0) return 0.;

// Find first column with least colcount
   for(k=p+1; k<=ncols; k++){
      if ( cost  == 1) break;     //preemptive
      if ((count = colcount(k)) < 1 ) continue;
      if ( count >= cost ) continue;
      p = k; cost = count;
   }
   if (p != r ) swapcols(r,p);

// Find max pivot value in this column
   elem = col[r-1];  val = 0.;  cost = ncols+1;  p = r;
   while ((k = elem->mext) != 0){ elem = e+k;
      if ((i = elem->row) < r) continue;     //Assert i >= r !
      if ((v = fabs(elem->value)) <= val ) continue;
      val = v;  p = i;  Urr = elem->value;  cost = rowcount(i); }

// Find row with acceptable value and least rowcount
   elem = col[r-1]; val *= THRESHOLD;
   while ((k = elem->mext) != 0){ elem = e+k;
      if ((i = elem->row) < r) continue;     //Assert i >= r !
      if ((count = rowcount(i)) >= cost || count < 1) continue;
      if ( fabs(elem->value) < val ) continue;
      p = i;  cost = count;  Urr = elem->value; }

   if (p != r) swaprows(r,p);
   return Urr;
}

void LDLt_test (const sparse& L, const real_vector& D)
{//evaluates L*D*Lt; caller sets L(i,i) = 1.
   unsigned i,j,k,n=D.N();  double aij;
   if(L.N() != n || L.M() != n)
     { printf("!!!!! Incompatible Dimensions !!!!!\n"); return; }
   for(i=1; i<=n; i++){
      for(j=1; j<=i; j++){   aij =  0.;
         for(k=1; k<=i; k++) aij += L.getval(i,k) *D(k)* L.getval(j,k);
         cout << '|' << aij; }
      cout << "|\n"; }
}

void Min_Deg_Pivot (sparse& A, pivot& P)
{//sorts A and corresponding P
   unsigned k, m, n, p, r, count, cost;
   if((m = A.M()) != P.m) exit(1);  if((n = A.N()) != P.n) exit(1);

   for(r=1; r<=m; r++)
   {  if ((cost = A.rowcount(r)) == 0 ) continue;  p = r;
//    Find first row with least rowcount
      for(k=p+1; k<=m; k++){
         if ((count = A.rowcount(k)) >= cost ) continue;
            p = k; cost = count; }
      if (p == r ) continue;  A.swaprows(r,p);  P.swaprows(r,p);
   }

   for(r=1; r<=n; r++)
   {  if ((cost = A.colcount(r)) == 0 ) continue;  p = r;
//    Find first col with least colcount
      for(k=p+1; k<=n; k++){
         if ((count = A.colcount(k)) >= cost ) continue;
            p = k; cost = count; }
      if (p == r ) continue;  A.swapcols(r,p);  P.swapcols(r,p);
   }
}

void Sym_Pivot (sparse& A, pivot& P)
{//sorts A and corresponding P
   unsigned k, m, p, r, count, cost;
   if((m = A.M()) != A.N() || m != P.n || m != P.m ) exit(1);

   for(r=1; r<=m; r++)
   {
      if ((cost = A.rowcount(r)) == 0 ) continue;  p = r;
//    Find first row with least rowcount
      for(k=p+1; k<=m; k++){
         if ((count = A.rowcount(k)) >= cost ) continue;
         p = k; cost = count;
      }
      if (p == r ) continue;
      A.swaprows(r,p); A.swapcols(r,p);
      P.swaprows(r,p); P.swapcols(r,p);
   }
}

void print_stats(const sparse& S)
{  unsigned r,i,j,k,maxr,maxc,minr,minc;
   maxc = maxr = i = j = 0;  minc = minr = S.NZ()+1;
   for(r=1; r<=S.M(); r++){
      if((k = S.rowcount(r)) > maxr){ maxr = k;  j = r; }
      if( k                  < minr){ minr = k;  i = r; }
   }
   printf("Row %u has %u, Row %u has %u", i, minr, j, maxr);
   i = j = 0;
   for(r=1; r<=S.N(); r++){
      if((k = S.colcount(r)) > maxc){ maxc = k;  j = r; }
      if( k                  < minc){ minc = k;  i = r; }
   }
   printf(", Col %u has %u, Col %u has %u\n", i, minc, j, maxc);
}

void print_test (const sparse& S)
{  int quit = 0; unsigned r;
   do{
      printf("\nEnter r <= %u: ", S.M()); scanf("%u",&r);
      if( r > 0 && r <= S.M() )
         printf("rowcount %u = %u colcount %u = %u\n",
         S.rowcount(r), S.rowpop(r,1), S.colcount(r), S.colpop(1,r));
      else quit = 1;
     } while(quit == 0);
}

void print_rows (const sparse& S)
{  int quit = 0; unsigned irow;
   do{
      printf("\nEnter row <= %u: ", S.M()); scanf("%u",&irow);
      if( irow > 0 && irow <= S.M() ) S.dumprow(irow);
      else quit = 1;
     } while(quit == 0);
}

void print_cols (const sparse& S)
{  int quit = 0; unsigned icol;
   do{
      printf("\nEnter col <= %u: ", S.N()); scanf("%u",&icol);
      if( icol > 0 && icol <= S.N() ) S.dumpcol(icol);
      else quit = 1;
     } while(quit == 0);
}

void print_matrix (const sparse& A)
{  unsigned i,j,m=A.M(),n=A.N();
   for(i=1; i<=m; i++){
      for(j=1; j<=n; j++) cout << '|' << A.getval(i,j);
                          cout << "|\n"; }
}
