/* sparse.h                                                     12/28/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/
#ifndef SPARSE_H_INCLUDED
#define SPARSE_H_INCLUDED
#include <stdio.h>
#include <math.h>
class matrix;

struct pivot
{
   unsigned* R; unsigned* C;
   unsigned m, n;

   pivot() { R = C = NULL;  m = n = 0;}
   pivot(unsigned, unsigned);
   void dumpall (unsigned = 0) const;
   void swaprows (const unsigned, const unsigned);
   void swapcols (const unsigned, const unsigned);
   void init();
};

struct element
{
   unsigned row, col;
   double value;
   unsigned next, mext;
};

class sparse
{
   friend class real_vector;
   friend class temp_real_vector;
   friend temp_real_vector operator*(const sparse&, const real_vector&);
   friend temp_real_vector operator*(const real_vector&, const sparse&);

private:

   element *e;
   unsigned mrows, ncols, size, kount;
   element** row;  element** col;

public:

   sparse();
   sparse(const unsigned, const unsigned, const unsigned);
   ~sparse();

   double& operator () (const unsigned, const unsigned); //Fortran style
   double& stash (const unsigned, const unsigned);       //Unconditional
   double getval (const unsigned, const unsigned) const; //read-only
   void   unlink   (const unsigned, const unsigned);
   void row_unlink (const unsigned, const unsigned);
   void col_unlink (const unsigned, const unsigned);
   void reset (const pivot&);
   double row_maxab (const unsigned);
   void row_scale (const unsigned, const double);
   void rowneg (const unsigned);
   element& nonz (const unsigned k) const {return e[mrows+ncols+k]; }
   unsigned M() const { return mrows; }
   unsigned N() const { return ncols; }
   unsigned NZ() const{ return kount-mrows-ncols; }
   unsigned& rowcount (const unsigned i) const{ return row[i-1]->col; }
   unsigned& colcount (const unsigned j) const{ return col[j-1]->row; }
   unsigned inspect() const;
   void dumprow (const unsigned) const;
   void dumpcol (const unsigned) const;
   void dumpall (const unsigned=0) const;
   void swaprows (const unsigned, const unsigned);
   void swapcols (const unsigned, const unsigned);
   unsigned rowpop (const unsigned, const unsigned) const;
   unsigned colpop (const unsigned, const unsigned) const;
   void LDLt     (const sparse&, real_vector&);
   void LDLt_QD  (const sparse&, real_vector&);
   void LDLt_QDt (const sparse&, real_vector&);
   void LDLt_QD  (const matrix&, real_vector&);
   void LDLt_QDt (const matrix&, real_vector&);
   unsigned Crout_LU (pivot&);
   unsigned Gauss_LU ();
   unsigned Gauss_LU (pivot&);
   double Pivot (const unsigned);
};

void LDLt_test  (const sparse&, const real_vector&);
void Min_Deg_Pivot(sparse&, pivot&);
void Sym_Pivot(sparse&, pivot&);
void print_stats(const sparse&);
void print_test (const sparse&);
void print_rows (const sparse&);
void print_cols (const sparse&);
void print_matrix (const sparse&);

//void checkptr(void*);

inline sparse::sparse()
{
   e = 0; row = 0; col = 0; mrows = 0; ncols = 0; size = 0; kount = 0;
}

#include <iostream.h>

inline ostream& operator << (ostream& s, const element& q)
{
    s << " [(" << q.row << "," << q.col << "): " << q.value << " -> "
             << q.next << " " << q.mext << "]"; return s;
}

istream& operator >> (istream&, element&);

#endif SPARSE_H

