/* vector.cpp                                                   12/28/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h> // for mem* and cout
#include "..\book\source\vector.h"
#include "..\book\source\matrix.h"
#include "..\book\source\sparse.h"

enum{NO,YES} DEBUG = NO;

double operator * (const real_vector& r, const real_vector& l)
{
   double sumprod = 0.0; unsigned i;
   if(r.n == l.n)
      for (i=0; i<r.n; i++) sumprod += *(r.ptr+i) * *(l.ptr+i);
   return sumprod;
}

temp_real_vector operator+(const real_vector& v1, const real_vector& v2)
{
   if (v1.n != v2.n) exit(1);
   real_vector s (v1.n);
   s.sum(v1,v2);
   return temp_real_vector(s);
}

temp_real_vector operator-(const real_vector& v1, const real_vector& v2)
{
   if (v1.n != v2.n) exit(1);
   real_vector s (v1.n);
   s.dif(v1,v2);
   return temp_real_vector(s);
}

temp_real_vector operator*(const real_vector& v1, const double c)
{
   real_vector s (v1.n);
   s.byc(v1,c);
   return temp_real_vector(s);
}

real_vector::real_vector(double * data, unsigned length, int stride)
{
   n = length;  ptr = new double  [n];  checkptr(ptr);
   for (unsigned i=0,j=0; i<n; i++,j+=stride) *(ptr+i) = *(data+j);
}

real_vector& real_vector::operator=(const real_vector& v1)
{
   if (ptr == NULL) {n = v1.n; ptr = new double [n]; checkptr(ptr);}
   memcpy(ptr, v1.ptr, n*sizeof(double));
   return *this;
}

real_vector& real_vector::operator*=(const double c)
{
   for(unsigned i=0; i<n; i++) *(ptr+i) *= c;
      return *this;
}

real_vector& real_vector::operator+=(const double c)
{
   for(unsigned i=0; i<n; i++) *(ptr+i) += c;
      return *this;
}

real_vector& real_vector::operator+=(const real_vector& v1)
{
   for(unsigned i=0; i<n; i++) *(ptr+i) += *(v1.ptr+i);
      return *this;
}

real_vector& real_vector::operator-=(const real_vector& v1)
{
   for(unsigned i=0; i<n; i++) *(ptr+i) -= *(v1.ptr+i);
      return *this;
}

real_vector& real_vector::operator- ()
{
   for(unsigned i=0; i<n; i++) ptr[i] = -ptr[i];
   return *this;
}

real_vector&
real_vector::sum( const real_vector& v1, const real_vector& v2)
{
   if (n != v1.n || n != v2.n) exit(1);
   for(unsigned i=0; i<n; i++) *(ptr+i) = *(v1.ptr+i) + *(v2.ptr+i);
   return *this;
}

real_vector&
real_vector::dif( const real_vector& v1, const real_vector& v2)
{
   if (n != v1.n || n != v2.n) exit(1);
   for(unsigned i=0; i<n; i++) *(ptr+i) = *(v1.ptr+i) - *(v2.ptr+i);
   return *this;
}

real_vector& real_vector::byc(const real_vector& v1, const double c)
{
   if (n != v1.n) exit(1);
   for(unsigned i=0; i<n; i++) *(ptr+i) = *(v1.ptr+i) * c;
   return *this;
}

real_vector& real_vector::init(unsigned len, double * buf)
{
   if (len != 0)
   {
      if (n != len)
      {
         if (ptr != NULL) exit(1);
         n = len; ptr = buf;
         if (buf != NULL) return *this;
         ptr = new double [n];  checkptr(ptr);
      }
   }
   if (n > 0) memset(ptr, 0, n*sizeof(double));
   return *this;
}

real_vector& real_vector::Lsolve(const sparse& s)
{//revised to columnwise access 12/29/97
   if(n != s.ncols) exit(1);
   element* elem;  unsigned i,j,k;  double x;
   for (j=1; j<n; j++)
   {  if((x = (*this)(j)) == 0.) continue;
      elem = s.col[j-1];
      while ((k = elem->mext) != 0){ elem = s.e+k;
         if ((i = elem->row)  >  j)
         (*this)(i) -= (elem->value) * x; }
   }
   return *this;
}

real_vector& real_vector::TLsolve(const sparse& s)
{
   if(n != s.mrows) exit(1);
   element* elem;  unsigned i,j,k;
   for (j=n-1; j>=1; j--)
   {
      elem = s.col[j-1];
      while ((k = elem->mext) != 0){ elem = s.e+k;
         if ((i = elem->row)  >  j)
         (*this)(j) -= (elem->value)*((*this)(i));}
   }
   return *this;
}

real_vector& real_vector::Usolve(const sparse& s)
{
   if(n != s.ncols) exit(1);
   element* elem;  unsigned i,j,k; double Uii;
   if ((Uii = s.getval(n,n)) == 0.) exit(1); (*this)(n) /= Uii;
   for (i=n-1; i>=1; i--)
   {
      if ((Uii = s.getval(i,i)) == 0.) exit(1);
      elem = s.row[i-1];
      while ((k = elem->next) != 0){ elem = s.e+k;
         if ((j = elem->col)  >  i)
         (*this)(i) -= (elem->value)*((*this)(j));}
      (*this)(i) /= Uii;
   }
   return *this;
}

real_vector& real_vector::TUsolve(const sparse& s)
{//revised to rowwise access 2/28/98
   if(n != s.mrows) exit(1);
   element* elem;  unsigned i,j,k; double Uii, x;
   if ((Uii = s.getval(1,1)) == 0.) exit(1); x = (ptr[0] /= Uii);
   for (i=1; i<n; i++)
   {
      if(x != 0.){
         elem = s.row[i-1];
         while ((k = elem->next) != 0){ elem = s.e+k;
            if ((j = elem->col)  >  i)
            ptr[j-1] -= (elem->value) * x; }
      }
      if ((Uii = s.getval(i+1,i+1)) == 0.) exit(1);
      x = (ptr[i] /= Uii);
   }
   return *this;
}

real_vector& real_vector::Qsolve (const matrix& A)
{//Apply Q stored as house vectors to b, rhs
   unsigned i, j; double alfa, beta;
   if(n != A.M()) exit(1);
   for (j=1; j<=A.N(); j++)
   {  beta = 1.; alfa = (*this)(j);
      for (i=n; i>j; i--)
      {
         beta += A(i,j)*A(i,j);     //v*v
         alfa += A(i,j)*(*this)(i); //v*b
      }
      alfa *= 2.; alfa /= beta;     //2vb/vv
      (*this)(j) -= alfa;           //b -= (2vb/vv)v
      for (i=n; i>j; i--) (*this)(i) -= A(i,j)*alfa;
   }
   return *this;
}

real_vector& real_vector::Lsolve (const matrix& A)
{//Solve Lx = b where L is lower triangular part of A
   unsigned i,j;  double x;
   for(j=1; j<A.N(); j++){
      if((x = (*this)(j)) == 0.) continue;
      for(i=j+1; i<=A.M(); i++) (*this)(i) -= A(i,j) * x;
   }
   return *this;
}

real_vector& real_vector::Usolve (const matrix& A)
{//Solve Rx = Qb where R is upper triangular part of A
   unsigned i, j, r; double Uii;
   r = A.N();
   for(i=r; i<n; i++) ptr[i] = 0.;
   if((Uii = A(r,r)) != 0.)(*this)(r) /= Uii;
      else                 (*this)(r) = 0.;
   for (i=r-1; i>0; i--)
   {
      for(j=i+1; j<=r; j++) (*this)(i) -= A(i,j)*(*this)(j);
      if((Uii = A(i,i)) != 0.)(*this)(i) /= Uii;
                         else (*this)(i) = 0.;
   }
   return *this;
}

real_vector& real_vector::Dsolve(const real_vector& d)
{//use after LDLt()
   if(n != d.n) exit(1);
   unsigned j;  double x;
   for (j=1; j<=n; j++)
   {  if((x = d(j)) == 0.){ (*this)(j) = 0.;  continue; }//theoretically
      (*this)(j) /= x;
   }
   return *this;
}

real_vector& real_vector::Map (const unsigned R[], double* buf)
{  unsigned q = 0;
   if (buf == NULL) { buf = new double [n]; checkptr(buf); q = n;}
   memcpy(buf, ptr, n*sizeof(double));
   for (unsigned i=0; i<n; i++) ptr[i] = *(buf+R[i]-1);
   if(q) delete buf;
   return *this;
}

real_vector& real_vector::Pam(const unsigned C[], double* buf)
{  unsigned q = 0;
   if (buf == NULL) { buf = new double [n]; checkptr(buf); q = n;}
   memcpy(buf, ptr, n*sizeof(double));
   for (unsigned i=0; i<n; i++) *(ptr+C[i]-1) = buf[i];
   if(q) delete buf;
   return *this;
}

void real_vector::dumpall (unsigned l) const
{
   unsigned j;
   if  (l == 0 ||l > n) l = n;
   for (j=0; j<l; j++) cout << "|" << ptr[j];
   cout << "\n";
}

temp_real_vector operator*(const sparse& A, const real_vector& b)
{
   if (A.N() != b.N()) exit(1);
   element* elem; unsigned i,k;
   real_vector s; s.init(A.M());
   for (i = 0; i < A.M(); i++) {  elem = A.row[i];
      while ((k = elem->next) != 0){ elem = A.e+k;
         s[i] += b(elem->col) * elem->value;}}
   return temp_real_vector(s);
}

temp_real_vector operator*(const real_vector& b, const sparse& A)
{
   if (A.M() != b.N()) exit(1);
   element* elem; unsigned j,k;
   real_vector s; s.init(A.N());
   for (j = 0; j < A.N(); j++) {  elem = A.col[j];
      while ((k = elem->mext) != 0){ elem = A.e+k;
         s[j] += b(elem->row) * elem->value;}}
   return temp_real_vector(s);
}

temp_real_vector operator*(const matrix& A, const real_vector& b)
{
   if (A.N() != b.N()) exit(1);
   real_vector s (A.M());
   for (unsigned i = 0; i < A.M(); i++) s[i] = A[i] * b;
   return temp_real_vector(s);
}

temp_real_vector operator*(const real_vector& b, const matrix& A)
{
   if (A.M() != b.N()) exit(1);
   unsigned i,j;
   real_vector s; s.init(A.N());
   for (i = 0; i< A.M(); i++)     // do i outer
      for (j = 0; j < A.N(); j++)
         s[j] += b[i] * A[i][j];  // s += (A[i]*b[i]);
   return temp_real_vector(s);
}

temp_real_vector house(const matrix& A, const unsigned j)
{//G&VL 5.1.1 p196
   real_vector s (A.M());
   for (unsigned i = 0; i < A.M(); i++) s[i] = A[i] (j);
   double mu = sqrt(s*s);
   if( mu != 0.){
      if(s(1) < 0.) mu = -mu; // s(1) > 0.? mu : mu = -mu;
      double beta = 1./(s(1) + mu); // 0 den is noop!
      s *= beta;}
   s(1) = 1.;
   return temp_real_vector(s);
}

temp_real_vector sym_house (matrix& A)
{//G&VL p196 applied to leading row & column (2:n)
   real_vector s (A.M()-1);
   for(unsigned i=0; i<s.N(); i++) s[i] = A[i+1] (1);
   double mu = sqrt(s*s);
   if( mu != 0.){
      if(s(1) < 0.) mu = -mu; // s(1) > 0.? mu : mu = -mu;
      double beta = 1./(s(1) + mu); // 0 den is noop!
      s *= beta;}
   s(1) = 1.;
   A(2,1) = A(1,2) = -mu;  // sub\superdiagonal
   for(unsigned i = 2; i <= s.N(); i++)        // stash s
      A[i][0] = A[0][i] = s(i);
   return temp_real_vector(s);
}

void checkptr(void* ptr)
{
   if (ptr == NULL) {
      printf("Insufficient memory for ptr \n");
      exit(1);}
   if (DEBUG) printf("new %Fp \n",ptr);
}
