/* vector.h                                                     12/28/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/
#ifndef VECTOR_H_INCLUDED
#define VECTOR_H_INCLUDED
#include <stdio.h>
#include <stdlib.h>

class real_vector;
class sparse;
class temp_real_vector;
class matrix;
class submatrix;
template<class T> struct cplxfy;

class real_vector
{
   friend class sparse;
   friend class temp_real_vector;
   friend class matrix;
   friend class submatrix;
   friend class cplxfy<real_vector>;
   friend double operator * (const real_vector&, const real_vector&);
   friend temp_real_vector operator+ (const real_vector&,
                                      const real_vector&);
   friend temp_real_vector operator- (const real_vector&,
                                      const real_vector&);
   friend temp_real_vector operator* (const real_vector&, const double);

protected:

   double  *ptr; unsigned n;

public:

   real_vector();
   real_vector(unsigned);
   real_vector(double *, unsigned);
   real_vector(double *, unsigned, int);

   double& operator [] (unsigned) const;
   double& operator () (unsigned) const;   // Fortran style
   unsigned N() const { return n; }
   double* zap() { delete ptr; return ptr; }
   real_vector& operator=(const real_vector&);
   real_vector& operator*=(const double);
   real_vector& operator+=(const double);
   real_vector& operator+=(const real_vector&);
   real_vector& operator-=(const real_vector&);
   real_vector& operator- ();
   real_vector& operator--() {       n--; return *this;} //MATRIX REQS
   real_vector& sum(const real_vector&, const real_vector&);
   real_vector& dif(const real_vector&, const real_vector&);
   real_vector& byc(const real_vector&, const double);
   real_vector& init(unsigned = 0, double * = NULL);
   real_vector& Lsolve(const sparse&);
   real_vector& TLsolve(const sparse&);
   real_vector& Usolve(const sparse&);
   real_vector& TUsolve(const sparse&);
   real_vector& Qsolve(const matrix&);
   real_vector& Lsolve(const matrix&);
   real_vector& Usolve(const matrix&);
   real_vector& Dsolve(const real_vector&);
   real_vector& Map(const unsigned[], double* = NULL);
   real_vector& Pam(const unsigned[], double* = NULL);
   real_vector& ift(cplxfy<real_vector>&);
   void swapc(unsigned, unsigned);
   void dumpall (unsigned = 0) const;
};

class temp_real_vector : public real_vector
{
private:

   temp_real_vector(const real_vector& t) { n = t.n;  ptr = t.ptr;}

public:

   ~temp_real_vector();
   friend temp_real_vector
   operator+(const real_vector&, const real_vector&);
   friend temp_real_vector
   operator-(const real_vector&, const real_vector&);
   friend temp_real_vector operator*(const real_vector&, const double);
   friend temp_real_vector
   operator*(const double a, const real_vector& v){ return v * a; }
   friend temp_real_vector operator*(const sparse&, const real_vector&);
   friend temp_real_vector operator*(const real_vector&, const sparse&);
   friend temp_real_vector operator*(const matrix&, const real_vector&);
   friend temp_real_vector operator*(const real_vector&, const matrix&);
   friend temp_real_vector house(const matrix&, const unsigned);
   friend temp_real_vector sym_house(matrix&);
};

void checkptr(void *);

#include <math.h>
inline double euc( const real_vector& x)
{
   return sqrt(x*x);
}

inline real_vector::real_vector()
{
   ptr = NULL; n = 0;
}

inline real_vector::real_vector(unsigned length)
{
   n = length;  ptr = new double  [n];  checkptr(ptr);
}

inline real_vector::real_vector(double * data, unsigned length)
{
   ptr = data; n = length;
}

inline double& real_vector::operator [] (unsigned i) const
{
   return ptr[i];
}

inline double& real_vector::operator () (unsigned i) const
{
   return ptr[i-1];
}

inline void real_vector::swapc (unsigned i, unsigned j)
{
   double t = ptr[i];  ptr[i] = ptr[j];  ptr[j] = t;  return;
}

inline temp_real_vector::~temp_real_vector()
{//                                 if (DEBUG) printf("del %Fp \n",ptr);
   delete ptr;
}

#endif VECTOR_H
