/* cplxfy.cpp                                                   12/28/00

     Copyright (c) John R. Berryhill 1999 - 2000.  All Rights Reserved.

*/
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include "..\book\source\vector.h"
#include "..\book\source\matrix.h"
#include "..\book\source\sparse.h"
#include "..\book\source\method.h"
#include "..\book\source\cplxfy.h"
#include <math.h>

cplxfy<real_vector>&
cplxfy<real_vector>::operator*= (const cplxfy<double>& c)
{//(z = z*c)
    real_vector tmp;//real_vector tmp(re->N());
    tmp = (*re)*(*c.re) - (*im)*(*c.im);
    *im = (*re)*(*c.im) + (*im)*(*c.re); *re = tmp;    tmp.zap();
    return *this;
}

cplxfy<real_vector>&
cplxfy<real_vector>::operator*= (const cplxfy<matrix>& C)
{//(z = C*z)
    real_vector tmp;
    tmp = (*C.re)*(*re) - (*C.im)*(*im);
    *im = (*C.im)*(*re) + (*C.re)*(*im); *re = tmp;    tmp.zap();
    return *this;
}

cplxfy<real_vector>&
cplxfy<real_vector>::operator*= (const real_vector& f)
{//real filter
   unsigned n;
   if( (n = re->n) != f.n ) exit(1);
   for(unsigned i=0; i<n; i++){
      (*re)[i] *= f[i];
      (*im)[i] *= f[i];
   }
   return *this;
}

cplxfy<real_vector>&
cplxfy<real_vector>::operator*= (const cplxfy<real_vector>& f)
{//complex filter
   double t; unsigned n;
   if( (n = re->n) != f.re->n ) exit(1);
   for(unsigned i=0; i<n; i++){
      t        = (*re)[i]*(*f.re)[i] - (*im)[i]*(*f.im)[i];
      (*im)[i] = (*re)[i]*(*f.im)[i] + (*im)[i]*(*f.re)[i];
      (*re)[i] = t;
   }
   return *this;
}

template<class T> double abs( const cplxfy<T>& a)
{
   double x = sqrt( real(a)*real(a) + imag(a)*imag(a) );
   return x;
}

temp_real_vector real(Csum& S)
{
   if (S.v1->re->N() != S.v2->re->N()) exit(1);
   real_vector s (S.v1->re->N());
   s.sum(real(*(S.v1)),real(*(S.v2)));
   return temp_real_vector(s);
}

temp_real_vector imag(Csum& S)
{
   if (S.v1->im->N() != S.v2->im->N()) exit(1);
   real_vector s (S.v1->im->N());
   s.sum(imag(*(S.v1)),imag(*(S.v2)));
   return temp_real_vector(s);
}

cplxfy<real_vector>& cplxfy<real_vector>::operator= (const Csum& rhs)
{
   *re = *(rhs.v1->re) + *(rhs.v2->re);
   *im = *(rhs.v1->im) + *(rhs.v2->im);
   return *this;
}

temp_real_vector real(Cdif& S)
{
   if (S.v1->re->N() != S.v2->re->N()) exit(1);
   real_vector s (S.v1->re->N());
   s.dif(real(*(S.v1)),real(*(S.v2)));
   return temp_real_vector(s);
}

temp_real_vector imag(Cdif& S)
{
   if (S.v1->im->N() != S.v2->im->N()) exit(1);
   real_vector s (S.v1->im->N());
   s.dif(imag(*(S.v1)),imag(*(S.v2)));
   return temp_real_vector(s);
}

cplxfy<real_vector>& cplxfy<real_vector>::operator= (const Cdif& rhs)
{
   *re = *(rhs.v1->re) - *(rhs.v2->re);
   *im = *(rhs.v1->im) - *(rhs.v2->im);
   return *this;
}

unsigned bitrev( const unsigned k, const unsigned log2_n)
{//VL92 1.5.1 p39
   unsigned s, m = k, j = 0;
   for(unsigned q=0; q<log2_n; q++){
      s = m/2; (j *= 2) += (m - 2*s); m = s; }
   return j;
}

unsigned next_pow_2( const unsigned length)
{
   if (length <= 1) return 1;
   unsigned n = exp(log2_n(length)*log(2.0)) + 0.5;
   return n;
}

unsigned sizeof_fft(const real_vector& input)
{
   unsigned N = next_pow_2(input.N()) / 2;
   return N + 1;
}

unsigned sizeof_fft(const cplxfy<real_vector>& input)
{
   unsigned N = next_pow_2(input.re->N());
   return N;
}

void cfft( cplxfy<real_vector>& z, unsigned log2_n, int sign)
{//FFT with bitrev preordering and without stored tables
   unsigned n = pow( 2, log2_n); unsigned i, j, m, k=1;
   double c, s, x, y, arg, sc, pi = 4.0 * atan(1.0);

   if(sign < 0) { sc = n;  sc = 1.0/sc;  z *= sc; }

// Bit-reverse preordering
   for(i=0; i<n; i++){
      j = bitrev( i, log2_n);
      if( j <= i ) continue;
      real(z).swapc( i, j);
      imag(z).swapc( i, j);
   }

// Gentleman-Sande algorithm [VL92 1.9.2 p67]
   while(k<n){
      for(m=0; m<k; m++){
         arg = pi * sign * m / k;
         c = cos(arg); s = sin(arg);
         for(i=m; i<n; i+=k*2){
            x            = real(z)[i+k]*c - imag(z)[i+k]*s;
            y            = real(z)[i+k]*s + imag(z)[i+k]*c;
            real(z)[i+k] = real(z)[i] - x;
            imag(z)[i+k] = imag(z)[i] - y;
            real(z)[i]  += x;
            imag(z)[i]  += y; 	}
      }
      k = k * 2;
   }
   return;
}

void cfft( cplxfy<matrix>& z, unsigned log2_n, int sign)
{//FFT with bitrev preordering and without stored tables
   unsigned n = pow( 2, log2_n); unsigned i, j, m, k=1;
   double c, s, arg, sc, pi = 4.0 * atan(1.0);
   unsigned N = real(z).N();  real_vector x(N), y(N);

   if(sign < 0){ sc = n; sc = 1.0/sc;
      for(i=0; i<n; i++){ real(z)[i] *= sc; imag(z)[i] *= sc; } }

// Bit-reverse preordering
   for(i=0; i<n; i++){
      j = bitrev( i, log2_n);
      if( j <= i ) continue;
      real(z).swaprows( i+1, j+1);
      imag(z).swaprows( i+1, j+1);
   }

// Gentleman-Sande algorithm [VL92 1.9.2 p67]
   while(k<n){
      for(m=0; m<k; m++){
         arg = pi * sign * m / k;
         c = cos(arg); s = sin(arg);
         for(i=m; i<n; i+=k*2){
            x            = real(z)[i+k]*c - imag(z)[i+k]*s;
            y            = real(z)[i+k]*s + imag(z)[i+k]*c;
            real(z)[i+k] = real(z)[i] - x;
            imag(z)[i+k] = imag(z)[i] - y;
            real(z)[i]  += x;
            imag(z)[i]  += y; 	}
      }
      k = k * 2;
   }
   x.zap(); y.zap();
   return;
}

void cfft( cplxfy<real_vector> z[], unsigned log2_n, int sign)
{//FFT with bitrev preordering and without stored tables
   unsigned n = pow(2,log2_n); unsigned i, j, m, k=1;
   double c, s, arg, sc, pi = 4.0 * atan(1.0);
   cplxfy<double> cw( c, s);
   cplxfy<real_vector> ct( real(z[0]).N() );

   if(sign < 0){ sc = n; sc = 1.0/sc; for(i=0; i<n; i++) z[i] *= sc; }

// Bit-reverse preordering
   for(i=0; i<n; i++){
      j = bitrev( i, log2_n);
      if( j <= i ) continue;
      swap( z[i], z[j]);
   }

// Gentleman-Sande algorithm [VL92 1.9.2 p67]
   while(k<n){
      for(m=0; m<k; m++){
         arg = pi * sign * m / k;
         real(cw) = -(cos(arg)); imag(cw) = -(sin(arg)); //now (-)
         for(i=m; i<n; i+=k*2){
            ct      = z[i];
            z[i+k] *= cw;     //was (-cw);
            z[i]   -= z[i+k];
            z[i+k] += ct;     }
      }
      k = k * 2;
   }
   real(ct).zap();  imag(ct).zap();  return;
}

void twofer( cplxfy<real_vector> z, unsigned log2_n, int sign)
{//Sort out or recombine FFTs of 2 real vectors
   unsigned i, n = pow(2,log2_n);
   double temp;
   for(i=1; i<n/2; i++){
      temp          = real(z)[n-i];
      real(z)[n-i]  = real(z)[i] - temp;
      real(z)[i]   += temp;
   }
   if(sign < 0) for(i=1; i<n/2; i++){
      temp          = imag(z)[n-i];
      imag(z)[n-i] += imag(z)[i];
      imag(z)[i]   -= temp;
   }
   else for(i=1; i<n/2; i++){
      temp          = imag(z)[i];
      imag(z)[i]   += imag(z)[n-i];
      imag(z)[n-i] -= temp;
   }
   if(sign < 0) for(i=1; i<n/2; i++){
      real(z)[i] *= .5;  real(z)[n-i] *= .5;
      imag(z)[i] *= .5;  imag(z)[n-i] *= .5;
   }
   return; //note that z[0] and z[n/2] are never touched
}

cplxfy<real_vector>& cplxfy<real_vector>::fft( const real_vector& in)
{//forward fft of real to complex
   unsigned i, n, nin, nyq, nhalf; nin = in.N(); nhalf = nin/2;
   nyq  = next_pow_2(nin)/2;  n = nyq + 1;
   if(!(n == re->n && n == im->n)) exit(1);

   for( i=0; i<nhalf; i++){
      (*re)[i] = in[2*i];  (*im)[i] = in[2*i+1]; }

   memset((re->ptr)+nhalf, 0, (n-nhalf)*sizeof(double));
   memset((im->ptr)+nhalf, 0, (n-nhalf)*sizeof(double));
   if( nin % 2 != 0) (*re)[nhalf] = in[nin-1];

   cfft( *this, log2_n(nyq), -1);
   (*re)[nyq] = (*re)[0];  (*im)[nyq] = (*im)[0];

   double x, y, a, b, u, v, s, c, p, q;
   p = -4.0*atan(1.0)/(double)nyq;
   q = sin(p);  p = cos(p);  c = 1.;  s = 0.;
   cplxfy<double> e(c,s), d(p,q);

   for (unsigned i=0; i<nyq/2; i++, e *= d) {
      a = 0.25*( (*re)[i] + (*re)[nyq-i]);
      b = 0.25*( (*im)[i] - (*im)[nyq-i]);
      x = 0.25*( (*re)[i] - (*re)[nyq-i]);
      y = 0.25*( (*im)[i] + (*im)[nyq-i]);

      u = c*x - s*y;         v = c*y + s*x;
      (*re)[i]     = a + v;  (*im)[i]     =   b - u;
      (*re)[nyq-i] = a - v;  (*im)[nyq-i] = -(b + u);
   }
   (*re)[nyq/2] *= 0.5; (*im)[nyq/2] *= -0.5;
   return *this;
}

real_vector& real_vector::ift( cplxfy<real_vector>& f)
{//inverse fft of complex f in place, return real
   unsigned i, nyq;
   nyq = real(f).N()-1;  if( nyq != next_pow_2(n)/2 ) exit(1);

   double x, y, a, b, u, v, s, c, p, q;
   p = 4.0*atan(1.0)/(double)nyq;
   q = sin(p);  p = cos(p);  c = 1.;  s = 0.;
   cplxfy<double> e(c,s), d(p,q);

   for( i=0; i<nyq/2; i++, e *= d){
      a =  real(f)[i] + real(f)[nyq-i];
      b =  imag(f)[i] - imag(f)[nyq-i];
      x =  real(f)[i] - real(f)[nyq-i];
      y =  imag(f)[i] + imag(f)[nyq-i];

      u = c*x - s*y;      v = c*y + s*x;
      real(f)[i]     = a - v;  imag(f)[i]     =   b + u;
      real(f)[nyq-i] = a + v;  imag(f)[nyq-i] = -(b - u);
   }
   real(f)[nyq/2] *= 2.;
   imag(f)[nyq/2] *= -2.;

   cfft( f, log2_n(nyq),  1);

   for( i=0; i<n; i+=2) *(ptr+i) = real(f)[i/2];
   for( i=1; i<n; i+=2) *(ptr+i) = imag(f)[i/2];
   if( n % 2 != 0)    *(ptr+n-1) = real(f)[n/2];
   return *this;
}

cplxfy<real_vector>&
cplxfy<real_vector>::phase_shift( int tshift, unsigned nyq)
{//tshift is in units of time-domain samples.
   unsigned i, n = re->n;
   double p, q, t, s = 0., c = 1.;
   q = -tshift*4.*atan(1.)/(double)nyq; //*(2*pi/n)
   p = sin(0.5*q); p = -2.*p*p; q = sin(q);
                                               //  -(*im);
   for(i=1; i<n; i++){
      t = c;  c += t*p - s*q;  s += t*q + s*p;
      t        = (*re)[i]*c - (*im)[i]*s;
      (*im)[i] = (*re)[i]*s + (*im)[i]*c;
      (*re)[i] = t;                       }    //  -(*im);
   return *this;
}

cplxfy<real_vector>&
cplxfy<real_vector>::bandpass( double  lo, double hi, unsigned nyq)
{//User initializes re[] to 1 and im[] to 0, if pattern
   unsigned i, j, n;
   if((n = re->n) < nyq+1 ) exit(1);  j = 2*nyq-1;
   double x, w, a=0., b=0.;
   double half_pi = 2.0*atan(1.0);
   double      bl = tan(lo*half_pi);
   double    finc = half_pi/nyq;    double freq = finc;
   if( hi < 1.0){ a = 1.0/tan(hi*half_pi);  b = a*bl;
                  (*re)[nyq] = 0.;   (*im)[nyq] = 0.; }
   if( lo > 0.0){ (*re)[0]   = 0.;   (*im)[0]   = 0.; }

   for(i=1; i<nyq; i++, j--, freq+=finc){
      x = tan(freq);
      x = (a*x*x-bl)/(x*(1.0-b));
      w = 1.0 / sqrt(1.0 + pow(x, 4.));  //4: 12db/octave
      (*re)[i] *= w;  (*im)[i] *= w;
      if(j<n){
      (*re)[j] *= w;  (*im)[j] *= w; }
   }
   return *this;
}

void phase_shift_2D( cplxfy<matrix>& z, int tshift)
{//tshift is in units of time-domain samples.
   unsigned i, j, m, n;  //     2-D grid is square
   n = real(z).N(); m = real(z).M();  if(n != (m/2+1)) exit(1);
   double c, p, q, s, t;
   p = tshift*8.*atan(1.)/(double)m; //tshift * (2 * pi / m)

   for(i=0; i<m/2; i++){ q = (double)i * (double)i;
      for(j=0; j<n; j++){
         t = p * sqrt( q + (double)j * (double)j );
         c = cos(t); s = sin(t);
         t             = real(z)[i][j]*c - imag(z)[i][j]*s;
         imag(z)[i][j] = real(z)[i][j]*s + imag(z)[i][j]*c;
         real(z)[i][j] = t;
      }
   }
   for(i=m/2; i<m; i++){ q = (double)(m-i) * (double)(m-i);
      for(j=0; j<n; j++){
         t = -p * sqrt( q + (double)j * (double)j );
         c = cos(t); s = sin(t);
         t             = real(z)[i][j]*c - imag(z)[i][j]*s;
         imag(z)[i][j] = real(z)[i][j]*s + imag(z)[i][j]*c;
         real(z)[i][j] = t;
      }
   }
   return;
}
